(*  Title:     strand
    ID:         $Id: strand.thy,v 1.00 2008/07/02 13:41:34 Yongjian Li ver1.0 $
    Author:     Yongjian Li, Institute of software, Chinese academy of sciences
    Copyright   2008  Institute of software, Chinese academy of sciences

Datatypes of strands, bundles, components, authentication tests;
the main results of strand space theories
*)

header{*Theory of strand spaces for Security Protocols*}

theory strand05=Event+Public:

section{*Main defintions*}
datatype
  Sign=positive  ("+"  100)
  |negative ("-" 100)



typedecl sigma

types
  signed_msg="Sign\<times>  msg"
types
 
  node=" sigma \<times> nat"

types
  strand_space="sigma   => signed_msg  list "


consts
  Sigma_set::"sigma set" ("\<Sigma>")

  SP::"strand_space"

constdefs
  Domain::"node set"
  "Domain == {(n1,i). n1 \<in>  \<Sigma>  \<and> i < length (SP n1)}"

constdefs strand::"node => sigma"
  "strand n==fst n"

constdefs index ::"node => nat"
  "index n == snd n"

constdefs node_sign ::"node => Sign"
   "node_sign n ==fst (nth (SP(fst(n))) (snd(n)) )"



constdefs 
  node_term ::"node\<Rightarrow>msg"
  "node_term n==  snd (nth (SP(fst(n))) (snd(n)) )"

constdefs

  casual1:: "( node  \<times> node ) set "  
  "casual1 == { (n1,n2) . n1 \<in>  Domain \<and> n2 \<in>  Domain \<and>
  node_sign n1= +  \<and>
  node_sign n2= -
  \<and>  node_term n1= node_term n2 
  \<and>  strand n1 \<noteq> strand n2    
  } "

 
syntax
  "_casual1"::" node \<Rightarrow>node\<Rightarrow>bool" (infix "\<rightarrow>" 100)

translations
  "n1\<rightarrow>n2 "=="(n1 ,n2) \<in> casual1"

 


constdefs
  casual2::"(node  \<times> node) set"
  "casual2 ==  { (n1,n2) . n1 \<in>  Domain \<and> n2 \<in>  Domain  \<and>
 
  (strand n1)=   (strand n2 ) \<and> Suc (index n1)=index n2} "
syntax
  "_casual2"::" node \<Rightarrow>node\<Rightarrow>bool" (infix "\<Rightarrow>" 50)

translations
  "n1\<Rightarrow>n2 "=="(n1 ,n2):casual2"

syntax
  "_casual2Trans"::" node \<Rightarrow>node\<Rightarrow>bool" (infix "\<Rightarrow>\<^sup>+" 50)

translations
  "n1\<Rightarrow>\<^sup>+n2 "=="(n1 ,n2):casual2^+"

syntax
  "_casual2TransReflex"::" node \<Rightarrow>node\<Rightarrow>bool" (infix "\<Rightarrow>\<^sup>*" 50)

translations
  "n1\<Rightarrow>\<^sup>*n2 "=="(n1 ,n2):casual2^*"

constdefs
  casual3::"(node  \<times> node) set"
  "casual3 ==  { (n1,n2) . n1\<rightarrow>n2 \<or>(n1, n2):(casual2^+) }"
syntax
  "_casual3"::" node \<Rightarrow>node\<Rightarrow>bool" (infix "\<mapsto> " 50)

translations
  "n1\<mapsto> n2 "=="(n1 ,n2):casual3"


types
  edge="node  \<times> node"
  graph="node set \<times> edge set"

consts 
  attr::"sigma\<Rightarrow> agent"
  (* "attr s== (fst (SP s))"*)






constdefs
  Atoms::"msg set"
  "Atoms=={a. (\<exists> ag. a=Agent ag) \<or>  (\<exists> n. a=Number n) 
  \<or>(\<exists> n. a=Nonce n) | (\<exists> k. a=Key k)}"
syntax
  Is_atom::"msg=> bool"
translations
  "Is_atom m"=="m  \<in>  Atoms"

consts KP::"key set"
defs KP_def:"KP=={k. \<exists> A. (k=pubK A)|(A \<in> bad \<and>(k=priK A |k=shrK A))}"
consts T:: "msg set"
defs T_is_only_text: "T=={t. t: Atoms\<and> (\<forall> k. t \<noteq>   Key k)}"


constdefs
  Is_K_strand::"sigma \<Rightarrow> bool "
  "Is_K_strand KS ==   (\<exists> k.  k\<in> KP \<and> (SP KS)=[(+, Key k)])"


constdefs
  Is_T_strand::"sigma \<Rightarrow> bool"
  "Is_T_strand KS==   (\<exists> t. t \<in> T  \<and>  (SP KS)=[(+,  t)]) "


constdefs
  Is_E_strand ::"sigma \<Rightarrow> bool "
  "Is_E_strand KS ==   (\<exists> k. \<exists> h.  (SP KS)=[(-, (Key k)),(-,h),(+, (Crypt k h))])"

constdefs
  Is_D_strand::"sigma \<Rightarrow>bool"
  "Is_D_strand KS==  (\<exists> k. \<exists> k'. \<exists> h. k'=invKey k\<and> (SP KS)=[(-, (Key k')),(-, (Crypt k h)),(+,h)])"
constdefs
  Is_Cat_strand::"sigma \<Rightarrow>bool"
  "Is_Cat_strand KS==  (\<exists> g. \<exists> h.  (SP KS)=[(-, g),(-,  h),(+, MPair g h)])"  

constdefs
  Is_Sep_strand::"sigma \<Rightarrow>bool"
  "Is_Sep_strand KS==  (\<exists> g. \<exists> h.  (SP KS)=[(-, MPair g h),(+,  g),(+,  h)])"


constdefs
  Is_Flush_strand::"sigma \<Rightarrow>bool"
  "Is_Flush_strand KS==  (\<exists> g.   (SP KS)=[(-, g )])"
constdefs
  Is_Tee_strand::"sigma \<Rightarrow>bool"
  "Is_Tee_strand KS==  (\<exists> g.  (SP KS)=[(-, g),(+,  g),(+,  g )])"  

consts Is_penetrator_strand:: "sigma => bool"
defs penetrator_trace: "Is_penetrator_strand s==
     (  Is_Tee_strand s | Is_Flush_strand s | Is_Cat_strand s |
        Is_Sep_strand s |  Is_E_strand s | Is_D_strand s |
        Is_T_strand s | Is_K_strand s)"

constdefs Is_regular_strand :: "sigma \<Rightarrow>bool"
  "Is_regular_strand s==\<not> ( Is_penetrator_strand s)"



constdefs nodes::"graph \<Rightarrow> node set"
  "nodes b== fst b"

constdefs edges::"graph \<Rightarrow> edge set"
  "edges b==snd b"


consts   bundles ::" graph set"
inductive "bundles"
  intros

  Nil: "( {} , {}) : bundles"
  Add_positive1: "\<lbrakk> b \<in>   bundles; 
  (node_sign n2) = +;
  n2 \<in> Domain;
  n2 \<notin> (nodes b);
  0 < index n2 ;
  n1 \<in>  nodes  b;
  n1\<Rightarrow> n2
  \<rbrakk>\<Longrightarrow>
  ({n2} Un (nodes b), {(n1, n2)} Un (edges b)) \<in> bundles"

  Add_positive2: "\<lbrakk> b \<in>   bundles; 
  node_sign n2=+;
  n2 \<notin> (nodes b);
  n2 \<in> Domain;
  index n2=0 
  \<rbrakk>\<Longrightarrow>
  ({n2} Un nodes b,  edges b) \<in> bundles"

  Add_negtive1: "[| b \<in>   bundles; 
  node_sign n2=-;
  n2 \<notin> nodes b;               
  ((strand  n1 \<noteq> strand n2)\<and> (n1 \<rightarrow> n2) \<and> (n1  \<in>  nodes b) \<and> (\<forall> n3. ( (n3  \<in>  nodes b)\<longrightarrow> (n1,n3) \<notin>   edges b)) );
  0 < index n2 ;
  n1' \<in>  nodes b;
  n1'\<Rightarrow>n2
  |]==>
  ({n2} Un nodes b,  {(n1, n2), (n1' , n2)} Un edges b) \<in> bundles"

  Add_negtive2: "[| b \<in>   bundles; 
  node_sign n2=-;
  n2 \<notin> nodes b;               
  ((strand  n1 \<noteq> strand n2)\<and> (n1 \<rightarrow> n2) \<and> (n1  \<in>  nodes b) \<and> (\<forall> n3. ( (n3  \<in>  nodes b)\<longrightarrow> (n1,n3) \<notin>   edges b)) );
  index n2=0
  |]==>
  ({n2} Un nodes b, {(n1, n2)} Un edges b) \<in> bundles"  

(*syntax
  "_edgebTrans"::" node\<Rightarrow>node\<Rightarrow>graph\<Rightarrow>bool" ( "("_ \<prec>\<^sub> _  _ "  ) " 

translations
  "n1\<prec>\<^sub>b n2 "=="(n1 ,n2):(edge b)^+"

syntax
  "_edgebTransReflex"::" node\<Rightarrow>node\<Rightarrow>graph\<Rightarrow>bool" (infix "\<preceq> \<^sub>b" 50)

translations
  "n1\<preceq> \<^sub>b n2 "=="(n1 ,n2):(edge b)^*"*)

constdefs
  parts_relation:: "msg\<Rightarrow>msg\<Rightarrow>bool"(infix "\<sqsubset> " 55)
  "parts_relation t1 t2==(t1  \<in>   (Message.parts {t2}))"
  
constdefs
  proper_subterm::"msg\<Rightarrow>msg\<Rightarrow>bool"
  "proper_subterm t1 t2 == t1 \<sqsubset> t2 \<and> t1\<noteq>t2"

constdefs 
  originate:: "msg\<Rightarrow>node\<Rightarrow>bool"
  "originate t n==(node_sign n)=+ \<and> t \<sqsubset> node_term n 
  \<and> (\<forall> i. i < index n \<longrightarrow> \<not> t \<sqsubset>  node_term (strand n,i))"

constdefs 
  unique_originate:: "msg\<Rightarrow>node\<Rightarrow>bool"
  "unique_originate t n== originate t n \<and> (\<forall> n'. originate t n'\<longrightarrow> n=n')"
constdefs 
  non_originate::" msg\<Rightarrow>sigma\<Rightarrow>bool"
  "non_originate t s==\<forall> n. (strand n=s\<longrightarrow> 
  (\<not>t \<sqsubset> node_term n| node_sign n=- | (\<exists> i. i< index n \<and> t \<sqsubset> node_term (strand n,i) )))" 

constdefs regularK::"nat \<Rightarrow> graph\<Rightarrow>bool"
  "regularK k b== \<forall> n\<in> (nodes b).(node_term n=Key k  )\<longrightarrow>Is_regular_strand (strand n)"

constdefs component::"msg \<Rightarrow> graph \<Rightarrow>bool"
  "component t b== (\<forall> g h. t\<noteq>(MPair g h)) \<and>
                   (\<forall> k h . (t=Crypt k h) \<longrightarrow> (regularK (invKey k) b))"

constdefs suite::"msg set \<Rightarrow>msg\<Rightarrow>graph \<Rightarrow>bool"
  "suite M a b==(\<forall> t\<in> M. (a \<sqsubset> t \<longrightarrow> component t b)) \<and>
                (\<forall> t. (\<not>a\<sqsubset> t \<longrightarrow> t\<in> M))"

constdefs
  path:: "msg\<Rightarrow>node list\<Rightarrow>graph\<Rightarrow>bool" 
  "path a p  b== a \<sqsubset> node_term ( p ! (length p -  1)) \<and> (
  \<forall> i. (i <length p - 1  \<longrightarrow> (  a \<sqsubset> node_term ( p ! i) \<and>
  ((p!i, p!(i+1)) \<in>  (edges b Int casual1) \<or> 
  ( (p!i, p!(i+1)) \<in>  (edges b Int casual2)^+ 
  ))))   )     "

constdefs
  orig_path:: "node list \<Rightarrow>graph \<Rightarrow>bool"
  "orig_path p b == p=[] | (\<exists> n. p=[n]\<and> n \<in>  (nodes b))  | (\<forall> i. (i <length p - 1  \<longrightarrow> (  (p!i, p!(i+1)) \<in>  (edges b Int casual1) \<or> 
  (p!i, p!(i+1)) \<in>  (edges b Int  casual2)^+ 
  )
  ))"
constdefs
  first_node_in_nonorigi_strand::"msg\<Rightarrow>node\<Rightarrow>node\<Rightarrow>bool"
  "first_node_in_nonorigi_strand a n m== node_sign m=- \<and>  a \<sqsubset> node_term m \<and> strand m=strand n \<and> (\<forall> i0< index m. \<not> a \<sqsubset> node_term  (strand m , i0))"

consts
  slice_arr_cons::"sigma\<Rightarrow> nat \<Rightarrow>nat \<Rightarrow>  node list"
primrec

  "slice_arr_cons s i (Suc len) =  (s,i  )# (slice_arr_cons s (Suc i)  len)" 
  "slice_arr_cons s j 0=[(s,j)]"

consts 
 complete_Path ::"msg  \<Rightarrow> graph\<Rightarrow>((node list) set)"
 inductive "complete_Path  a b"
 intros 
    (*uniqueOriginate: "unique_originate a m \<Longrightarrow> [m] \<in> (complete_Path  a b)"*)
    addPos1:"\<lbrakk> non_originate  a (strand m2);
             node_sign m2=+; a\<sqsubset>node_term  m2;
             first_node_in_nonorigi_strand a m2 m;
             p \<in> complete_Path a b ; last p=m             \<rbrakk>
          \<Longrightarrow>p@(slice_arr_cons (strand m) (index m) (index m2 - index m)) \<in> complete_Path a b"

    addPos2:"\<lbrakk> unique_originate a m; strand m2=strand m;a\<sqsubset>node_term  m2;
              node_sign m2=+ \<rbrakk>
          \<Longrightarrow> (slice_arr_cons (strand m) (index m) (index m2 - index m)) \<in> complete_Path a b"   

    addnege: "\<lbrakk>m1\<rightarrow>m2; m1 \<in> nodes b; (m1,m2) \<in> edges b;
               p \<in> complete_Path a b; a\<sqsubset>node_term  m2; last p=m1\<rbrakk>
          \<Longrightarrow>(p@[m2]) \<in>  complete_Path a b"
    

syntax
  "_casualBundleTrancl"::" node \<Rightarrow>graph \<Rightarrow>node\<Rightarrow> bool" ("(_ \<prec>\<^sub>_ _)" [70,65,65] 61)

translations
  "n1\<prec>\<^sub>b n2 "=="(n1 ,n2):(edges b)^+"

syntax
  "_casualBundleTranclRefl"::" node \<Rightarrow>graph \<Rightarrow>node\<Rightarrow> bool" ("(_ \<preceq>\<^sub>_ _)" [70,65,65] 61)

translations
  "n1\<preceq>\<^sub>b n2 "=="(n1 ,n2):(edges b)^*"
            
   
section{*basic lemmas*}
subsection{*basic lemmas on causal relations*}  
lemma casual2_rel_more:" \<lbrakk>n1\<Rightarrow>n2\<rbrakk>\<Longrightarrow>n2=(strand n1, (index n1) + 1)"
  apply (unfold strand_def index_def casual2_def)
  apply simp
  done

lemma casual2_aux1 :" strand n=a \<and> index n=b \<Longrightarrow>n=(a,b) "
  apply (unfold strand_def index_def,auto)
  done

lemma casual2_rel_less:" \<lbrakk>n1\<Rightarrow>n2\<rbrakk>\<Longrightarrow>n1=(strand n2, (snd n2) - 1)"
  apply (unfold strand_def index_def casual2_def)
  apply (simp split:nat_diff_split)
  apply auto
  apply (rule casual2_aux1)
  apply (unfold strand_def index_def,simp)
  done

lemma casual2_trancl_less: "m\<Rightarrow>\<^sup>+m'\<Longrightarrow> index m < index m' \<and>strand m=strand m'"
  apply( erule trancl_induct)
  apply(unfold casual2_def,unfold index_def, auto)
  done

lemma casual2_rtrancl1: "\<lbrakk>n2 \<in>  Domain; i<=index n2\<rbrakk> \<Longrightarrow>
  (strand n2,index n2 - i) \<in> Domain \<and> ((strand n2,index n2 - i)\<Rightarrow>\<^sup>* n2)"   (is " \<lbrakk>?A;?B i\<rbrakk>\<Longrightarrow>?P i")
proof(unfold strand_def index_def , induct i)
  case 0 show ?case by auto
  
next
  fix n
  assume IH:"\<lbrakk>n2 \<in> Domain; n \<le> snd n2\<rbrakk> \<Longrightarrow> (fst n2, snd n2 - n) \<in> Domain \<and> ((fst n2, snd n2 - n)\<Rightarrow>\<^sup>* n2) "
    and a1: " n2 \<in> Domain" and a2:"Suc n \<le> snd n2"
  show    "(fst n2, snd n2 - Suc n) \<in> Domain \<and> ((fst n2, snd n2 - Suc n)\<Rightarrow>\<^sup>* n2)  " 
   (is "?P' n") 
  proof -
    from a2 have B4:" n \<le> snd n2 " by auto
    from this and IH and a1  have B5: "?P n" 
      by(unfold strand_def index_def,auto)
    from B5 and a1 and a2  
    have B7:"(fst n2, snd n2 - Suc n) \<in>  Domain\<and>((fst n2, snd n2 - Suc n), (fst n2, snd n2 -n)) \<in> casual2"
      apply-
      apply(unfold casual2_def, unfold Domain_def strand_def index_def) apply auto
      apply arith apply arith
      done
    from this and B5   show "?P' ( n)" 
      by (unfold strand_def index_def, auto intro:  converse_rtrancl_into_rtrancl) 
  qed
qed 

lemma casual2_rtrancl2:"\<lbrakk>n2 \<in>  Domain; i<=index n2\<rbrakk> \<Longrightarrow>
  (strand n2,  i) \<in> Domain \<and> ((strand n2, i)\<Rightarrow>\<^sup>* n2) "
  apply(unfold  strand_def index_def)
  apply(subgoal_tac "snd n2 - i<=snd n2") prefer 2 apply simp
  apply(subgoal_tac "snd n2 - (snd n2 - i) =i") prefer 2 apply simp
  apply(fold  strand_def index_def)
  apply(drule_tac i="snd n2 - i" in casual2_rtrancl1)
  apply(unfold  strand_def index_def) apply auto  done

lemma  casual2_trancl:"\<lbrakk>n2 \<in>  Domain; i<snd n2\<rbrakk> \<Longrightarrow>
  (fst n2,  i) \<in> Domain \<and> ((fst n2, i) \<Rightarrow>\<^sup>+n2)"
  thm casual2_rtrancl2 
  apply(drule_tac i="i" and ?n2.0="n2" in  casual2_rtrancl2) 
  apply(unfold  strand_def index_def, simp)
  apply(erule conjE, drule  rtranclD ) apply(erule disjE) prefer 2 apply blast 
  apply(cut_tac surjective_pairing[where p="n2"])
  apply(drule_tac s="n2" and t="(fst n2, snd n2)" and P="\<lambda> s.(fst n2, i) = s" in subst)
  apply simp apply( subgoal_tac "i=snd n2") 
  apply simp apply(thin_tac "(fst n2, i) = n2")
  apply (simp only:Pair_eq)  done





lemma  casual1_casual2_disjoint : "\<lbrakk>n1\<rightarrow>n2 \<and> n3\<Rightarrow>n4\<rbrakk> \<Longrightarrow> n1\<noteq>n3 \<or> n2\<noteq>n4"
  apply(unfold casual1_def, unfold casual2_def)
  apply(auto )
  done

lemma send_node_is_plus: "\<lbrakk>(n1\<rightarrow>n2)\<rbrakk>\<Longrightarrow> node_sign n1=+ \<and> n1 \<in> Domain"
  apply( unfold node_sign_def , unfold casual1_def, unfold Domain_def)
  apply(unfold node_sign_def node_term_def strand_def,auto )
  done

lemma recv_node_is_minus: "\<lbrakk>(n2\<rightarrow>n1)\<rbrakk>\<Longrightarrow> node_sign n1=- \<and> n1 \<in> Domain"
  apply(unfold casual1_def, unfold node_sign_def, unfold Domain_def)

  apply( auto)
  done



lemma casual2_rel_n1_n2_less:" \<lbrakk>n1\<Rightarrow>n2\<rbrakk>\<Longrightarrow>
  (strand n1= strand n2) \<and> index n1= (index n2) - 1"
  apply (unfold casual2_def index_def strand_def)
  apply (simp split:nat_diff_split)
 
  done

lemma casual_diff:" \<lbrakk>n1\<rightarrow> n2 | n1\<Rightarrow>n2\<rbrakk> \<Longrightarrow> n1\<noteq>n2"
  apply(unfold casual1_def, unfold casual2_def)
  apply( auto dest:recv_node_is_minus casual2_rel_less)
  done

lemma is_casual2 :
  "\<lbrakk> m:Domain; n \<in> Domain; strand  n= strand m ;  index n= Suc (index m)\<rbrakk>\<Longrightarrow>m\<Rightarrow>n"
  apply(unfold casual2_def ,unfold Domain_def)
  apply auto
  done

lemma casual1_judge_sign : "\<lbrakk>n1 \<rightarrow> n2\<rbrakk> \<Longrightarrow>strand n1 \<noteq>strand n2"
  apply (unfold casual1_def, auto)
  done

lemma casual2_judge_sign : "\<lbrakk>n1 \<Rightarrow> n2\<rbrakk> \<Longrightarrow>strand n1 = strand n2"
  apply (unfold casual2_def, auto)
  done

lemma  casul2_trancl_sign: 
  "\<lbrakk>(n1 , n2): casual2^+\<rbrakk> \<Longrightarrow>strand n2 = strand n1 \<and> (index n1 < index n2)"
  apply(erule trancl_trans_induct)
  apply(unfold casual2_def)
  apply auto
  done


lemma node_equal : " (n=n')= (strand n=strand n' \<and> index n=index n')"
  apply(unfold strand_def index_def)
  apply(simp add:Pair_fst_snd_eq)
  done


lemma node_sign_is_Plus_or_minus:" ((n ::Sign)= + | n=-)"
  (*apply(rule allI)*)
  apply(case_tac n)
  apply blast+

  done

lemma rtrancl_mono1:" \<lbrakk>x:r^*; r\<subseteq> s\<rbrakk>\<Longrightarrow> x \<in>  s^*"
  apply(blast dest:rtrancl_mono) done

lemma part_MPair:" \<lbrakk>(a\<noteq>{|g,h|}); a \<sqsubset> {|g,h|}\<rbrakk>\<Longrightarrow> a \<sqsubset> g | a \<sqsubset> h"
  apply(unfold parts_relation_def, simp)
  apply(simp only:parts_insert2)
  apply blast
  done

subsection{*basic lemmas on subterm relation*}
lemma subterm_atom:"\<lbrakk>a \<sqsubset> t; t \<in> Atoms \<rbrakk>\<Longrightarrow> a=t"
  apply(unfold Atoms_def)
  apply(simp, erule disjE) apply(unfold parts_relation_def) 
  apply(erule exE,simp  add:parts_insert_Agent)
  apply(erule disjE,erule exE,simp  add:parts_insert_Number)
  apply(erule disjE,erule exE,simp  add:parts_insert_Nonce)
  apply(erule exE,simp  add:parts_insert_Key)
  done



lemma subterm_trans:" \<lbrakk>g \<sqsubset> h; h \<sqsubset> l\<rbrakk> \<Longrightarrow> g \<sqsubset> l"
  
  apply(unfold parts_relation_def) thm parts_mono 
  apply( blast dest:parts_trans)  done

lemma subterm_itself:" h \<sqsubset> h"
  apply(unfold parts_relation_def) 
  apply(rule_tac H="{h}" and  X="h" in  parts.Inj) 
  apply simp done

lemma  body_subterm:"t'' \<sqsubset>  Crypt k h \<Longrightarrow> t'' =  Crypt k h | t'' \<sqsubset>   h"
  apply(unfold parts_relation_def) 
  apply(erule parts.induct)
  apply blast
  apply(fold parts_relation_def) apply(subgoal_tac "X \<sqsubset> {|X,Y|}") prefer 2 
  apply(unfold parts_relation_def,blast intro:parts.Fst parts.Snd  parts.Body,fold parts_relation_def)
  apply(erule disjE,simp, blast  dest:subterm_trans)
  apply(subgoal_tac "Y \<sqsubset> {|X,Y|}") prefer 2 
  apply(unfold parts_relation_def,blast intro:parts.Fst parts.Snd  parts.Body,fold parts_relation_def)
  apply(erule disjE,simp, blast  dest:subterm_trans)
  apply(subgoal_tac "X \<sqsubset> Crypt K X") prefer 2 
  apply(unfold parts_relation_def,blast intro:parts.Fst parts.Snd  parts.Body,fold parts_relation_def)
  apply(erule disjE,simp) prefer 2 apply( blast  dest:subterm_trans)
  apply(blast intro:subterm_itself)
  done

section{*the soundness of the inductive bundle model*}
    
lemma bundle_node_in_Domain: assumes a1:"b \<in> bundles"  shows c:"n \<in> nodes b \<longrightarrow> n \<in> Domain"

  using a1
proof( induct)
  case Nil  show ?case by (unfold nodes_def,auto)
  
next  

  fix b n1 n2
  assume a1: " b \<in> bundles" and a2:" n \<in> nodes b \<longrightarrow> n \<in> Domain" and
    a3:" node_sign n2 = +" and a4:" n2 \<in> Domain"  and a5:"n2 \<notin> nodes b"
    and a6:"0 < index n2" and a7:"n1 \<in> nodes b" and a8:" n1 \<Rightarrow> n2"
  show   " n \<in> nodes ({n2} \<union> nodes b, {(n1, n2)} \<union> edges b) \<longrightarrow> n \<in> Domain"
  proof -
    
    from a1 and a2 and a3 and a4 and a5 and a6 and a7 and a8
    show ?thesis apply -   apply(unfold nodes_def edges_def,rule impI) 
      apply( simp only:fst_conv)
      apply simp apply blast done
  qed

next
  
  fix b n1 n2
  assume a1: " b \<in> bundles" and a2:" n \<in> nodes b \<longrightarrow> n \<in> Domain" and
    a3:" node_sign n2 = +" and a4:" n2 \<in> Domain"  and a5:"n2 \<notin> nodes b"
    and a6:" index n2=0" 
  show   " n \<in> nodes  ({n2} \<union> nodes b,  edges b) \<longrightarrow> n \<in> Domain"
  proof -
    
    from a1 and a2 and a3 and a4 and a5 and a6 
    show ?thesis apply -   apply(unfold nodes_def edges_def,rule impI) 
      apply( simp only:fst_conv)
      apply simp apply blast done
  qed

next
  fix b n1 n1' n2
  assume a1: " b \<in> bundles" and a2:" n \<in> nodes b \<longrightarrow> n \<in> Domain" and
    a3:" node_sign n2 = -"   and a5:"n2 \<notin> nodes b"
    and a6:" 0<index n2" 
    and a7:"strand n1 \<noteq> strand n2 \<and> n1 \<rightarrow> n2 \<and> n1 \<in> nodes b \<and>
            (\<forall>n3. n3 \<in> nodes b \<longrightarrow> (n1, n3) \<notin> edges b)"
    and a8:"n1' \<in> nodes b" and a9:" n1' \<Rightarrow> n2"
  show   " n \<in> nodes ({n2} \<union> nodes b, {(n1, n2), (n1', n2)} Un  edges b) \<longrightarrow> n \<in> Domain"
  proof -
    
    from a1 and a2 and a3 and a5 and a6  and a7 and a8 and a9
    show ?thesis 
      apply -   apply(unfold nodes_def edges_def strand_def,rule impI) 
      apply( simp only:fst_conv)
      apply((erule conjE)+, unfold casual1_def, simp) 
      apply blast done
  qed

next
  fix b n1  n2
  assume a1: " b \<in> bundles" and a2:" n \<in> nodes b \<longrightarrow> n \<in> Domain" and
    a3:" node_sign n2 = -"   and a5:"n2 \<notin> nodes b"
    and a6:" index n2=0"  
    and a7:"strand n1 \<noteq> strand n2 \<and> n1 \<rightarrow> n2 \<and> n1 \<in> nodes b 
            \<and> (\<forall>n3. n3 \<in> nodes b \<longrightarrow> (n1, n3) \<notin> edges b)"
  show   " n \<in> nodes ({n2} \<union> nodes b, {(n1, n2)} Un  edges b) \<longrightarrow> n \<in> Domain"
  proof -
    
    from a1 and a2 and a3 and a5 and a6  and a7
    show ?thesis 
      apply -   
      apply(unfold nodes_def edges_def strand_def,rule impI) 
      apply( simp only:fst_conv)
      apply((erule conjE)+, unfold casual1_def, simp) 
      apply blast done
  qed
qed

lemma bundle_nodes_strand_in_space: assumes a1:"b \<in> bundles" 
  and a2:"m \<in> nodes b"

shows "strand m \<in>  \<Sigma>"
  apply -
  apply(cut_tac a1 a2,drule_tac n="m" in  bundle_node_in_Domain)
  apply(simp)
  apply(unfold Domain_def strand_def,auto)
  done
  




inductive_cases pair_is_a_bundle [elim!]: "(a,b) \<in>  bundles"

lemma bundle_edge_is_exist:
"\<lbrakk>b \<in>  bundles \<rbrakk>\<Longrightarrow> 
 ((node_sign n2=-) \<and> n2  \<in>  (nodes b))\<longrightarrow>(\<exists> n1. (n1\<rightarrow>n2) \<and> n1 \<in>  (nodes b) \<and>(n1, n2) \<in> (edges b))"    
  thm bundles.induct
  apply(erule bundles.induct, unfold nodes_def edges_def strand_def index_def)

  apply simp   (*goal 1*)

  apply (rule impI)
  apply ( simp only:fst_conv snd_conv)
  apply(simp)
  apply(erule conjE)
  apply(erule disjE)
  apply simp
  apply simp
  apply(erule exE)+
  thm exI
  apply (rule_tac ?x="a" in exI)
  apply (rule_tac ?x="ba" in exI)
  apply simp   (*goal 2*)

  apply (rule impI)
  apply ( simp only:fst_conv snd_conv)
  apply(simp)
  apply(erule conjE)
  apply(erule disjE)
  apply simp
  apply simp
  apply(erule exE)+
  thm exI
  apply (rule_tac ?x="a" in exI)
  apply (rule_tac ?x="ba" in exI)
  apply simp   (*goal 3*)

  apply (rule impI)
  apply ( simp only:fst_conv snd_conv)

  apply(erule conjE)+
  apply simp
  apply(erule disjE)
  apply (rule_tac ?x="fst n1" in exI)

  apply (rule_tac ?x="snd n1" in exI)
  apply simp

  apply simp
  thm exE
  apply(erule exE)+
  thm exI
  apply (rule_tac ?x="a" in exI)
  apply (rule_tac ?x="ba" in exI)
  apply simp (*goal 4*)

  apply (rule impI)
  apply ( simp only:fst_conv snd_conv)

  apply(erule conjE)+
  apply simp
  apply(erule disjE)
  apply (rule_tac ?x="fst n1" in exI)

  apply (rule_tac ?x="snd n1" in exI)
  apply simp
  apply simp
  thm exE
  apply(erule exE)+
  thm exI
  apply (rule_tac ?x="a" in exI)
  apply (rule_tac ?x="ba" in exI)
  apply simp

  done

   
lemma bundle_is_cl_under_edge: 
  "\<lbrakk>b:bundles\<rbrakk>\<Longrightarrow>(n1,n2) \<in> edges b \<longrightarrow>n1 \<in> nodes b \<and> n2 \<in> nodes b"
  apply(erule bundles.induct,unfold nodes_def edges_def)
  apply auto
  done

lemma  bundle_is_cl_under_edge1: 
  "\<lbrakk>b \<in> bundles;(n1,n2) \<in> edges b\<rbrakk>\<Longrightarrow>n1 \<in> nodes b \<and> n2 \<in> nodes b"

  apply(auto dest!: bundle_is_cl_under_edge)
  done

lemma bundle_is_not_close_edge: 
  "\<lbrakk>b \<in> bundles\<rbrakk>\<Longrightarrow>n1:nodes b \<and> n2\<notin>nodes b \<longrightarrow>(n1, n2)\<notin>edges b"
  apply(rule impI)
  apply(rule notI)
  apply(auto dest: bundle_is_cl_under_edge)
  done

lemma bundle_is_not_close_edge1:
  "\<lbrakk>b \<in> bundles;n1 \<in> nodes b ; n2\<notin>nodes b\<rbrakk>\<Longrightarrow>(n1, n2)\<notin>edges b"
  apply(auto dest: bundle_is_not_close_edge)
  done



lemma bundle_is_closed:"\<lbrakk>b \<in> bundles\<rbrakk>\<Longrightarrow> ((y, z) \<in>  edges b)\<longrightarrow>y  \<in> nodes b \<and> z \<in> nodes b"
  apply(erule bundles.induct, unfold nodes_def edges_def)
  apply (auto dest:send_node_is_plus recv_node_is_minus )
  done 

lemma bundle_is_closed1:"\<lbrakk>b  \<in> bundles\<rbrakk>\<Longrightarrow> ((y, z) \<in>  edges b)\<longrightarrow>y \<noteq> z"
  apply(erule bundles.induct,unfold  edges_def)
  apply(auto)

  done 


lemma bundle_casual1_prec_is_unique: 
  assumes A:"b \<in>  bundles "  
  shows "(node_sign n2=-) \<and>  n1\<rightarrow>n2 \<and> n1' \<rightarrow> n2\<and>
         (n1, n2) \<in> (edges b)\<and>(n1', n2) \<in> (edges b)\<longrightarrow>n1=n1'"
proof -
  have Goal0:"node_sign n2 = - \<and> n1 \<rightarrow> n2 \<and> n1' \<rightarrow> n2 
              \<and> (n1, n2) \<in> edges b \<and> (n1', n2) \<in> edges b
              \<and> n2 \<in> nodes b \<longrightarrow> n1 = n1'"
    
    using A
  proof( induct )
    case Nil show ?case 
      by(unfold nodes_def,  auto)
  next
    
    fix "b" "n1a" "n2a"          
    assume "b \<in> bundles"  
      and b:" node_sign n2 = - \<and> n1 \<rightarrow> n2 \<and> n1' \<rightarrow> n2 \<and> 
         (n1, n2) \<in> edges b \<and> (n1', n2) \<in> edges b \<and> n2 \<in> nodes b \<longrightarrow> n1 = n1'"
      and a:" node_sign n2a = +" and 
      " n2a \<in> Domain" and " n2a \<notin>nodes b" 
      and " 0 < index n2a" and " n1a \<in>nodes b" and " n1a \<Rightarrow> n2a"
    show "node_sign n2 = - \<and>
      n1 \<rightarrow> n2 \<and>
      n1' \<rightarrow> n2 \<and>
      (n1, n2) \<in> edges ({n2a} \<union>nodes b, {(n1a, n2a)} \<union> edges b) \<and>
      (n1', n2) \<in> edges ({n2a} \<union>nodes b, {(n1a, n2a)} \<union> edges b) \<and> 
      n2 \<in> nodes ({n2a} \<union>nodes b, {(n1a, n2a)} \<union> edges b) \<longrightarrow>
      n1 = n1'"
    proof  (unfold edges_def nodes_def,rule impI)
      assume c0:"node_sign n2 = - \<and>
        n1 \<rightarrow> n2 \<and>
        n1' \<rightarrow> n2 \<and>
        (n1, n2) \<in> snd ({n2a} \<union> fst b, {(n1a, n2a)} \<union> snd b) \<and>
        (n1', n2) \<in> snd ({n2a} \<union> fst b, {(n1a, n2a)} \<union> snd b) \<and>
    n2 \<in> fst ({n2a} \<union> fst b, {(n1a, n2a)} \<union> snd b)"
      show "n1=n1'"
      proof (cases "n2=n2a") 
    case True
        assume  c:"n2=n2a" show ?thesis 
        proof (rule FalseE)
          from a and c and c0 show False by auto
            (*from a and c  have d:"node_sign n2=+ "    by simp 
            from c0  have e: "node_sign n2= - " by simp
            from e and d show False by simp*)
        qed
      next
    case False
        assume  c:"n2 \<noteq>n2a" 
        show "n1=n1'"
        proof - 
          from c0  have "n2 \<in> fst ({n2a} \<union> fst b, {(n1a, n2a)} \<union> snd b)" by simp      
          then have f:"n2 : {n2a} Un fst b" by simp
          from c and f have g0':" n2 \<in> fst b" by simp
          from c0  have g:"(n1, n2) \<in> snd ({n2a} \<union> fst b, {(n1a, n2a)} \<union> snd b)" by simp 
          from g and c have g0:"(n1, n2) \<in> snd b" by simp 
          from c0 and c  have g':"(n1', n2) \<in> snd b" by simp 
          from c0 and g0 and g' and g0'  and b show ?thesis 
        by (unfold edges_def nodes_def,simp)
        qed
      qed
    qed
  next 
    fix "b" "n1a" "n2a"          
    assume "b \<in> bundles" and 
      b:" node_sign n2 = - \<and> n1 \<rightarrow> n2 \<and> n1' \<rightarrow> n2 \<and>
          (n1, n2) \<in> edges b \<and> (n1', n2) \<in> edges b \<and> n2 \<in> nodes b \<longrightarrow> n1 = n1'"
      and a:" node_sign n2a = +" and " n2a \<in> Domain" 
      and " n2a \<notin> nodes b" and " index n2a =0" 
    show "node_sign n2 = - \<and>
      n1 \<rightarrow> n2 \<and>
      n1' \<rightarrow> n2 \<and>
      (n1, n2) \<in> edges ({n2a} \<union> nodes b, edges b) \<and>
      (n1', n2) \<in> edges ({n2a} \<union> nodes b, edges b) \<and> 
      n2 \<in> nodes ({n2a} \<union> nodes b, edges b) \<longrightarrow>
      n1 = n1'"
    proof(unfold edges_def nodes_def,rule impI)
      assume c0:"node_sign n2 = - \<and>
    n1 \<rightarrow> n2 \<and>
    n1' \<rightarrow> n2 \<and>
    (n1, n2) \<in> snd ({n2a} \<union> fst b,snd b) \<and>
    (n1', n2) \<in> snd ({n2a} \<union> fst b,snd b) \<and> n2 \<in> fst ({n2a} \<union> fst b,snd b)"
      show "n1=n1'"
      proof (cases "n2=n2a") 
    case True
        assume  c:"n2=n2a" show ?thesis 
        proof (rule FalseE)
          from a and c and c0 show False by auto
            (*from a and c  have d:"node_sign n2=+ "    by simp 
            from c0  have e: "node_sign n2= - " by simp
            from e and d show False by simp*)
        qed
      next
    case False
        assume  c:"n2 \<noteq>n2a" 
        show "n1=n1'"
        proof - 
          from c0  have "n2 \<in> fst ({n2a} \<union> fst b, snd b)" by simp      
          then have f:"n2 : {n2a} Un fst b" by simp
          from c and f have g0':" n2:fst b" by simp
          from c0  have g:"(n1, n2)  \<in>  snd b" by simp 
          from g and c have g0:"(n1, n2) \<in>snd b" by simp 
          from c0 and c  have g':"(n1', n2) \<in> snd b" by simp 
          from c0 and g and g0' and b show ?thesis  
        by(unfold nodes_def edges_def, simp)
        qed
      qed
    qed
  next
    fix "b" "n1a" "n1'a" "n2a"
    assume 
      bb:"b \<in> bundles" and
      IH:" node_sign n2 = - \<and> n1 \<rightarrow> n2 \<and> n1' \<rightarrow> n2 \<and>
          (n1, n2) \<in> edges b \<and> (n1', n2) \<in> edges b \<and> n2 \<in> nodes b
          \<longrightarrow> n1 = n1'" and
      " node_sign n2a = -" and as0:"n2a \<notin> nodes b" 
      and  n1ab: "strand n1a \<noteq> strand n2a \<and> n1a \<rightarrow> n2a \<and> n1a \<in> nodes b
                  \<and> (\<forall>n3. n3 \<in> nodes b \<longrightarrow> (n1a, n3) \<notin> edges b)"
      and  " 0 < index n2a" and " n1'a \<in> nodes b" and n1'ab:"n1'a \<Rightarrow> n2a"

    show "(node_sign n2 = - \<and>
      n1 \<rightarrow> n2 \<and>
      n1' \<rightarrow> n2 \<and>
      (n1, n2) \<in> edges ({n2a} \<union> nodes b, {(n1a, n2a), (n1'a, n2a)} \<union> edges b) \<and>
      (n1', n2) \<in> edges ({n2a} \<union> nodes b, {(n1a, n2a), (n1'a, n2a)} \<union> edges b) \<and>
      n2 \<in> nodes ({n2a} \<union> nodes b, {(n1a, n2a), (n1'a, n2a)} \<union> edges b) 
      \<longrightarrow>  n1 = n1')" 
      
    proof(rule impI, unfold nodes_def edges_def index_def strand_def)
      from n1ab have ass0: "n1a\<rightarrow>n2a" by simp
      assume 
        as1: " node_sign n2 = - \<and>
        n1 \<rightarrow> n2 \<and>
        n1' \<rightarrow> n2 \<and>
        (n1, n2) \<in> snd ({n2a} \<union> fst b, {(n1a, n2a), (n1'a, n2a)} \<union> snd b) \<and>
        (n1', n2) \<in> snd ({n2a} \<union> fst b, {(n1a, n2a), (n1'a, n2a)} \<union> snd b) \<and>
        n2 \<in> fst ({n2a} \<union> fst b, {(n1a, n2a), (n1'a, n2a)} \<union> snd b)"

      show "n1=n1'" 
      proof (cases "n2=n2a")
        case True
        from this have ascase: "n2=n2a" .
        from as1  have as4:"(n1, n2) \<in> {(n1a, n2a), (n1'a, n2a)} \<union> snd b"  by simp
        from as1 have  as5:"(n1', n2) \<in>  {(n1a, n2a), (n1'a, n2a)} \<union> snd b" by simp
        from bb and as0 and ascase and as4  
    have as6:"(n1, n2) \<in> {(n1a, n2a), (n1'a, n2a)}"
          apply simp  apply(erule disjE,simp)+
          thm  bundle_is_closed 
      apply(fold nodes_def edges_def, drule_tac y="n1 " and z="n2a" in bundle_is_closed)
          apply simp done
        from as1 have as7: "n1 \<rightarrow> n2" by simp
        from as6 and as0 and n1'ab  and as7  have as8:"n1=n1a"
          apply simp  apply(erule disjE) apply simp 
      apply(blast dest:casual1_judge_sign casual2_judge_sign) done
        
        from bb and as0 and ascase and as5  
    have as6:"(n1', n2) \<in> {(n1a, n2a), (n1'a, n2a)}"
          apply simp  apply(erule disjE,simp)+  
      apply(fold nodes_def edges_def,drule_tac y="n1' " and z="n2a" in bundle_is_closed)
          apply simp done
        from as1 have as9: "n1'\<rightarrow> n2" by simp
        from as6 and as0 and n1'ab  and as9  have as10: "n1'=n1a"
          apply simp  apply(erule disjE) apply simp 
      apply(blast dest:casual1_judge_sign casual2_judge_sign) done
        from as8 and as10 show ?thesis  by simp
      next 
    case False
        assume  c:"n2 \<noteq>n2a" 
        show "n1=n1'"
        proof - 
          from as1  have "n2 \<in> fst ({n2a} \<union> fst b,  snd b)" by simp      
          then have f:"n2  \<in>  {n2a} Un fst b" by simp
          from c and f have g0':" n2 \<in> fst b" by simp
          from as1 and c have g:"(n1, n2)  \<in>  snd b" by simp 
          from as1 and c  have g':"(n1', n2) \<in> snd b" by simp 
          from as1 and g and g' and g0' and IH show ?thesis  
        by (unfold nodes_def edges_def,simp)
        qed     
      qed
    qed

  next
    fix "b" "n1a"  "n2a"
    assume 
      bb:"b \<in> bundles" and
      IH:" node_sign n2 = - \<and> n1 \<rightarrow> n2 \<and> n1' \<rightarrow> n2 \<and> (n1, n2) \<in> edges b 
          \<and> (n1', n2) \<in> edges b \<and> n2 \<in> nodes b \<longrightarrow> n1 = n1'" and
      " node_sign n2a = -" and as0:"n2a \<notin> nodes b" and
      n1ab: "strand n1a \<noteq> strand n2a \<and> n1a \<rightarrow> n2a \<and> n1a \<in> nodes b 
             \<and> (\<forall>n3. n3 \<in> nodes b \<longrightarrow> (n1a, n3) \<notin> edges b)"
      and       "index n2a=0" 

    show " node_sign n2 = - \<and>
      n1 \<rightarrow> n2 \<and>
      n1' \<rightarrow> n2 \<and>
      (n1, n2) \<in> edges ({n2a} \<union> nodes b, {(n1a, n2a)} \<union> edges b) \<and>
      (n1', n2) \<in>edges ({n2a} \<union> nodes b, {(n1a, n2a)} \<union> edges b) \<and> 
      n2 \<in> nodes ({n2a} \<union> nodes b, {(n1a, n2a)} \<union> edges b) \<longrightarrow>
      n1 = n1'" 
      
    proof(rule impI, unfold nodes_def edges_def index_def strand_def)
      from n1ab have ass0: "n1a\<rightarrow>n2a" by simp
      assume 
        as1: "node_sign n2 = - \<and>
        n1 \<rightarrow> n2 \<and>
        n1' \<rightarrow> n2 \<and>
        (n1, n2) \<in> snd ({n2a} \<union> fst b, {(n1a, n2a)} \<union> snd b) \<and>
        (n1', n2) \<in> snd ({n2a} \<union> fst b, {(n1a, n2a)} \<union> snd b) \<and> n2 \<in> fst ({n2a} \<union> fst b, {(n1a, n2a)} \<union> snd b)"

      show "n1=n1'" 
      proof (cases "n2=n2a")
        case True
        from this have ascase: "n2=n2a" .
        from as1  have as4:"(n1, n2) \<in> {(n1a, n2a)} \<union> snd b"  by simp
        from as1 have  as5:"(n1', n2) \<in>  {(n1a, n2a)} \<union> snd b" by simp
        from bb and as0 and ascase and as4  
    have as6:"(n1, n2) \<in> {(n1a, n2a)}"
          apply simp  apply(erule disjE,simp)+
      apply(fold nodes_def edges_def)
      apply(drule_tac y="n1 " and z="n2a" in bundle_is_closed)
          apply simp done
        from as1 have as7: "n1 \<rightarrow> n2" by simp
        from as6 and as0   and as7  have as8:"n1=n1a"
          apply simp  done
        
        from bb and as0 and ascase and as5  
    have as6:"(n1', n2) \<in> {(n1a, n2a)}"
          apply simp  apply(erule disjE,simp)+  
      apply(fold nodes_def edges_def)
      apply(drule_tac y="n1' " and z="n2a" in bundle_is_closed)
          apply simp done
        from as1 have as9: "n1'\<rightarrow> n2" by simp
        from as6 and as0  and as9  have as10: "n1'=n1a"
          apply simp done
        from as8 and as10 show ?thesis  by simp
      next 
    case False
        assume  c:"n2 \<noteq>n2a" 
        show "n1=n1'"
        proof - 
          from as1  have "n2 \<in> fst ({n2a} \<union> fst b,  snd b)" by simp      
          then have f:"n2 : {n2a} Un fst b" by simp
          from c and f have g0':" n2 \<in> fst b" by simp
          from as1 and c have g:"(n1, n2)  \<in>  snd b" by simp 
          from as1 and c  have g':"(n1', n2) \<in> snd b" by simp 
          from as1 and g and g' and g0' and IH show ?thesis  
        by (unfold nodes_def edges_def,simp)
        qed     
      qed
    qed
  qed
  show " node_sign n2 = - \<and> n1 \<rightarrow> n2 \<and> n1' \<rightarrow> n2 \<and> (n1, n2) \<in> edges b \<and>
         (n1', n2) \<in> edges b \<longrightarrow> n1 = n1'"
  proof(rule impI,unfold edges_def nodes_def) 
    assume b:"node_sign n2 = - \<and> n1 \<rightarrow> n2 \<and> n1' \<rightarrow> n2 \<and> (n1, n2) \<in> snd b \<and> (n1', n2) \<in> snd b" 
    show "n1=n1'"
    proof -
      from b have "(n1', n2) \<in> snd b" by simp
      from A and this have "n2 \<in> fst b" thm bundle_is_cl_under_edge1 
    apply(fold edges_def nodes_def)
    apply(drule_tac b="b" and ?n1.0="n1'" and ?n2.0="n2" in  bundle_is_cl_under_edge1)
    apply auto done
      from this and Goal0 and b show ?thesis 
    by  (unfold nodes_def edges_def,auto)
    qed
  qed
qed





lemma bundle_casual2_prec_is_unique:
  "\<lbrakk>b: bundles \<rbrakk>\<Longrightarrow> (  n1\<Rightarrow>n2 \<and> n2 \<in> nodes b )\<longrightarrow>((n1,n2) \<in> (edges b)
  \<and> n1 \<in> nodes b)" 
  apply(erule bundles.induct,unfold nodes_def edges_def )
  apply simp
  apply(rule impI,  simp only:fst_conv snd_conv)
  apply(erule conjE)+ apply(subgoal_tac "n2 \<in> {n2a} | n2 \<in>  fst b")
  prefer 2 apply blast
  apply(erule disjE) apply( simp)
  apply(subgoal_tac "n1=n1a") apply simp 
  apply(unfold casual2_def strand_def index_def,simp add:node_equal)
  apply(unfold strand_def index_def,simp)
  apply blast


  apply(rule impI,  simp only:fst_conv snd_conv )
  apply(subgoal_tac "n2a\<noteq>n2") apply blast
  apply(rule ccontr) apply  simp

  apply(rule impI,  simp only:fst_conv snd_conv)
  apply(erule conjE)+ apply(subgoal_tac "n2 \<in> {n2a} | n2 \<in>  fst b")
  prefer 2 apply blast
  apply(erule disjE) apply( simp)
  apply(subgoal_tac "n1=n1'") apply simp 
  apply(simp add:node_equal)
  apply(unfold strand_def index_def,simp)
  apply blast

  apply(rule impI,  simp only:fst_conv snd_conv) 
  apply(subgoal_tac "n2a\<noteq>n2") apply blast
  apply(rule ccontr) apply  simp 

  done

lemma bundle_node_is_in_Domains:"\<lbrakk>b \<in> bundles\<rbrakk>\<Longrightarrow>n \<in>  nodes b\<longrightarrow>n \<in> Domain"
  apply(erule bundles.induct)
  apply(unfold nodes_def casual1_def) apply auto
  done

lemma bundle_casual2_prec_is_unique_intro1:
  "\<lbrakk>b \<in>  bundles ;  n1\<Rightarrow>n2 ; n2 \<in> nodes b\<rbrakk>\<Longrightarrow>
  n1 \<in> (nodes b) \<and> n1 \<in> Domain " 

  apply(subgoal_tac " n1 \<in> nodes b") apply simp
  apply(blast dest:bundle_node_is_in_Domains)  
  apply(blast dest:bundle_casual2_prec_is_unique  )


  done

lemma bundle_casual2_closed_aux1:
  "\<lbrakk>0< index n2 - n; n2 \<in> Domain\<rbrakk>
  \<Longrightarrow> (strand n2, (index  n2) - Suc n) \<Rightarrow>(strand n2, (index n2) - n)"
  apply(unfold index_def strand_def casual2_def Domain_def)
  apply (auto ) 
  apply arith+
  done


lemma bundle_casual2_closed: 
  "\<lbrakk>b \<in> bundles;n2 \<in>  nodes b;  i\<le> index n2\<rbrakk>\<Longrightarrow>
  (strand n2, index n2 - i) \<in>  nodes b \<and> (strand n2, index n2 - i) \<in> Domain  "

  apply(induct i) thm bundle_node_is_in_Domains
  apply(unfold index_def strand_def)
  apply(force dest: bundle_node_is_in_Domains)
  apply(fold index_def strand_def) thm bundle_casual2_prec_is_unique_intro1
  apply(rule_tac ?b="b" and 
       ?n1.0="(strand  n2, index n2 - Suc i)"
        and ?n2.0="(strand n2, index n2 - i)"
    in  bundle_casual2_prec_is_unique_intro1)
  apply simp 
  apply(rule   bundle_casual2_closed_aux1)
  apply arith 
  apply(force dest: bundle_node_is_in_Domains)
  apply simp
  done





lemma bundle_casual2_closed2: "\<lbrakk>b \<in> bundles; n2 \<in>  nodes b;  i\<le>index  n2\<rbrakk>\<Longrightarrow>
  ( (strand n2,  i) \<in>  nodes b)\<and> (strand n2, i) \<in> Domain"
  apply(frule_tac i="index n2 - i"  in bundle_casual2_closed)
  apply simp+
  done

lemma bundle_casual2_closed222: 
 shows "\<lbrakk>b \<in> bundles; n2 \<in>  nodes b;  i\<le>index  n2\<rbrakk>
  \<Longrightarrow>((strand n2,index n2 - i), n2) \<in> (edges b \<inter> casual2)\<^sup>*"
  (is "\<lbrakk>?A1; ?A2;?A3\<rbrakk> \<Longrightarrow> ?P i")
proof(induct i )
  show "?P 0" 
    by(unfold nodes_def strand_def edges_def index_def,simp)
next
  fix n
  assume IH:"\<lbrakk>b \<in> bundles; n2 \<in> nodes b; n \<le> index n2\<rbrakk> \<Longrightarrow>?P n" and 
    B1:"b \<in> bundles" and B2:" n2 \<in> nodes b" and B3:" Suc n \<le> index n2"
  show "?P(Suc n)"
  proof -
    from B3 have B4:" n \<le> index n2 " by auto
    from this and IH and B1 and B2 have B5: "?P n" 
      by auto
    from B2 and this and B1  
    have B6: "(strand n2, index n2 - n) \<in> nodes b \<and> (strand n2, index n2 - n) \<in> Domain " 
      apply - apply(drule_tac i="index n2 - n" and ?n2.0="n2" and b="b" in bundle_casual2_closed2)
      apply auto done
    from B3 and B4 and B6 
    have B7:"((strand n2, index n2 - Suc n), (strand n2, index n2 -n)) \<in> casual2"
      apply-
      apply(unfold casual2_def, unfold Domain_def) apply auto
      apply arith 
      apply(unfold strand_def,simp) 
      apply(unfold index_def) apply simp 
      apply arith
      done
    from this and B1 and B6 
    have "((strand n2, index n2 - Suc n), (strand n2, index n2 -n)) \<in> edges b"
      by (auto dest:bundle_casual2_prec_is_unique)
    from this and B7 
    have B8:"((strand n2, index n2 - Suc n), (strand n2, index n2 -n)) \<in> (edges b Int casual2)" 
      by auto
    from B5 and this show "?P (Suc n)" by ( auto intro: converse_rtrancl_into_rtrancl) 
  qed
qed 



lemma bundle_casual2_closed3: 
  assumes A1 : "b \<in> bundles" and  A2:" n2 \<in>  nodes b" and  A3:" i\<le>index n2" (is "?A i")
  shows   "(strand n2,  i) \<in>  nodes b \<and> ( (strand n2,  i) ,n2): (edges b Int casual2)^*"
proof(rule conjI)
  from A1  and A2  and  A3
  show " (strand n2, i) \<in> nodes b"
    apply - apply(blast dest:bundle_casual2_closed2) done
next
  from A3  have C2:"index n2 - i <= index n2"
    by(unfold index_def, arith)
  from A1 and A2  and  A3 and C2 show "((strand n2,   i), n2) \<in> (edges b \<inter> casual2)\<^sup>*"  
    apply -
    thm bundle_casual2_closed222
    apply(drule_tac b="b" and ?n2.0="n2" and i="index n2 - i" in  bundle_casual2_closed222) 
    apply simp apply simp  apply auto done
qed


lemma bundle_casual2_trancl: 
" \<lbrakk>b \<in> bundles; n' \<in> (nodes b);  (m\<Rightarrow>\<^sup>+  n')\<rbrakk>\<Longrightarrow>m \<in>  (nodes b) \<and>(m\<prec>\<^sub>b n')"
  apply -
  apply(subgoal_tac "index m < index n' \<and>strand m=strand n'")
  prefer 2
  apply(blast dest:casual2_trancl_less)
  apply(subgoal_tac "index m <= index n' ")
  prefer 2
  apply simp
  apply(subgoal_tac "(strand n',  index m) \<in>  nodes b \<and> 
    ( (strand n',  index m) ,n') \<in>  (edges b Int casual2)^*")
  prefer 2
  apply(blast dest:bundle_casual2_closed3)
  apply(subgoal_tac "(strand n', index m)=m")
  apply simp
  apply(subgoal_tac "(m, n') \<in> (edges b)^*")
  apply(drule rtranclD) apply(erule disjE) apply simp
  apply simp
  apply(subgoal_tac "(edges b \<inter> casual2)\<^sup>* \<subseteq> (edges b)\<^sup>*")
  apply blast
  apply(blast intro!:rtrancl_mono) 
  apply(simp add:node_equal)
  apply(unfold strand_def index_def,auto)
  done

lemma bundle_edge_is_unique:
  "\<lbrakk>b \<in>  bundles \<rbrakk>\<Longrightarrow> 
  ((node_sign n2=-) \<and> n2  \<in>  (nodes b))\<longrightarrow>
  (\<exists>! n1. (n1\<rightarrow>n2) \<and> n1 \<in>  (nodes b) \<and>(n1, n2) \<in> (edges b))"    
  thm bundles.induct
    (*apply(erule bundles.induct)*)
  apply(blast dest: bundle_edge_is_exist bundle_casual1_prec_is_unique )
  done



lemma bundle_is_trans_closed2:
  "\<lbrakk>b \<in> bundles \<rbrakk>\<Longrightarrow>((y, z) \<in>  (edges b)^+)\<longrightarrow>y  \<in>  nodes b \<and> z \<in> nodes b"
  apply(rule impI)
    (*apply(erule rtrancl_induct)*)
  apply( erule trancl_trans_induct)

  apply (auto dest:bundle_is_closed bundle_is_closed1   )
  done 

lemma bundle_is_trans_closed22:
  "\<lbrakk>b \<in> bundles; y \<prec>\<^sub>b z \<rbrakk>\<Longrightarrow>y  \<in>  nodes b \<and> z \<in> nodes b"

  apply (auto dest:bundle_is_trans_closed2)
  done 

lemma bundle_is_rtrans_closed23:
  "\<lbrakk>b \<in> bundles; z \<in>  nodes b; (y \<preceq>\<^sub>b z)\<rbrakk>\<Longrightarrow>y  \<in>  nodes b \<and> z \<in> nodes b"
  apply(drule rtranclD)
  apply(auto dest: bundle_is_trans_closed22)
  done 

lemma bundle_is_rtrans_closed24:
  "\<lbrakk>b \<in> bundles; z\<notin>  nodes b;(y \<preceq>\<^sub>b z) \<rbrakk>\<Longrightarrow>y = z"
  apply(drule rtranclD)
  apply(auto dest: bundle_is_trans_closed22)
  done 

thm trancl_insert

lemma bundle_edge_is_anti2:
  assumes A:"b \<in> bundles" shows "y \<prec>\<^sub>b z\<longrightarrow> y\<noteq>z"
  using A
proof induct
  case Nil show ?case 
    by(unfold edges_def, auto)
next
  fix b n1 n2
  assume a1:"b \<in> bundles" 
    and IH:" (y, z) \<in> (edges b)\<^sup>+ \<longrightarrow> y \<noteq> z"
    and a2:" node_sign n2 = +" 
    and a3:" n2 \<in> Domain" and a4:" n2 \<notin> nodes b"
    and a5:" 0 < index n2" and a6:" n1 \<in> nodes b" and a7:" n1 \<Rightarrow> n2"
  show "(y, z) \<in> (edges ({n2} \<union> nodes b, {(n1, n2)} \<union> edges b))\<^sup>+ \<longrightarrow> y\<noteq>z"
  proof
    assume a8:"(y, z) \<in> (edges ({n2} \<union> nodes b, {(n1, n2)} \<union> edges b))\<^sup>+"
    show "y\<noteq>z"
    proof(cases "z=n2")
      case True
      from this have casehyp:"z=n2" .
      from a8 have a9:"(y,z):({(n1, n2)} \<union> edges b)\<^sup>+" 
    by(unfold edges_def, simp)
      from a9 have a10:"(y,z) \<in>  ((edges b)^+ Un {(a,c). (a,n1) \<in>  (edges b)^* \<and>  ((n2,c) \<in> (edges b)^*)})" 
    by (simp add:trancl_insert)
      from a10  have a12:"(y,z) \<in>  (edges b)^+ Un {(a,n2). (a,n1) \<in>  (edges b)^* }"   
      proof - 
    show "(y, z) \<in> (edges b)\<^sup>+ \<union> {(a, c). (a, n1) \<in> (edges b)\<^sup>* \<and> (n2, c) \<in> (edges b)\<^sup>*} 
      \<Longrightarrow> (y, z) \<in> (edges b)\<^sup>+ \<union> {(a, n2). (a, n1) \<in> (edges b)\<^sup>*}" apply simp
          apply(erule disjE) apply simp apply simp done
      qed
      from a12 and a6 and a4 and IH and casehyp and A  show ?thesis
      proof - 
    show "  \<lbrakk>(y, z) \<in> (edges b)\<^sup>+ \<union> {(a, n2). (a, n1) \<in> (edges b)\<^sup>*}; n1 \<in> nodes b;
             n2 \<notin> nodes b; (y, z) \<in> (edges b)\<^sup>+ \<longrightarrow> y \<noteq> z; z = n2; b \<in> bundles\<rbrakk>
      \<Longrightarrow> y \<noteq> z"
          apply simp
          apply(erule disjE) apply simp
          apply(frule_tac b="b" and y="y" and z="n1" in bundle_is_rtrans_closed23)
      apply auto   done
      qed
    next
      case False
      from this have casehyp:"z\<noteq>n2" .
      from  a8 have a9:"(y,z) \<in> ({(n1, n2)} \<union> edges b)\<^sup>+" 
    by (unfold edges_def,simp)
      from A and  a8 and casehyp and a4 
      have a10: "(y,z) \<in>  (edges b)\<^sup>+" 
      proof -  
        show "\<lbrakk>b \<in> bundles; z \<noteq> n2; (y, z) \<in> (edges ({n2} \<union> nodes b, {(n1, n2)} \<union> edges b))\<^sup>+; 
           n2 \<notin> nodes b\<rbrakk> \<Longrightarrow> (y, z) \<in> (edges b)\<^sup>+ " 
          apply (simp add:edges_def trancl_insert)
          apply(erule disjE) apply simp
          apply(erule conjE)
          apply(drule_tac a="n2" and b="z" in rtranclD)
      apply(fold edges_def)
          apply(auto dest:bundle_is_trans_closed22)
          done
      qed
      from a10 and IH show ?thesis by simp
    qed
  qed
next 
  fix b n2 
  assume a1: "b \<in> bundles" and a2: "(y, z) \<in> (edges b)\<^sup>+ \<longrightarrow> y \<noteq> z" 
         and a3:" node_sign n2 = +" and a4:" n2 \<notin> nodes b" 
         and a5:" index n2 = 0"
  show "(y, z) \<in> (edges ({n2} \<union> nodes b, edges b))\<^sup>+ \<longrightarrow> y \<noteq> z"
  proof
    assume a6:"(y, z) \<in> (edges ({n2} \<union> nodes b, edges b))\<^sup>+ "
    show "y\<noteq>z"
    proof -
      from a6 have a7:"(y, z) \<in> (edges b)^+" 
    by(unfold edges_def, auto)
      from a2 and a6 show ?thesis 
    by(unfold edges_def, simp)
    qed
  qed
next
  fix b n1 n1'  n2
  assume a1:"b \<in> bundles" 
    and IH:" (y, z) \<in> (edges b)\<^sup>+ \<longrightarrow> y \<noteq> z"
    and a2:" node_sign n2 = -" 
    and a3:" strand n1 \<noteq> strand n2 \<and> n1 \<rightarrow> n2 \<and> n1 \<in> nodes b \<and> (\<forall>n3. n3 \<in> nodes b \<longrightarrow> (n1, n3) \<notin> edges b)" 
    and a4:" n2 \<notin> nodes b" and a5:" 0 < index n2" and a6:" n1' \<in> nodes b" and a7:" n1' \<Rightarrow> n2"
  show "(y, z) \<in> (edges ({n2} \<union> nodes b, {(n1, n2), (n1', n2)} \<union> edges b))\<^sup>+ \<longrightarrow> y \<noteq> z"
  proof
    assume a8:"(y, z) \<in> (edges ({n2} \<union> nodes b, {(n1, n2),(n1',n2)} \<union> edges b))\<^sup>+"
    show "y\<noteq>z"
    proof(cases "z=n2")
      case True
      from this have casehyp:"z=n2" .
      from a8 have a9:"(y,z) \<in> ({(n1, n2),(n1',n2)} \<union> edges b)\<^sup>+" 
    by(unfold  edges_def,simp)
      from a9 
      have a10:"(y,z) \<in>  (({(n1',n2)} Un edges b)^+ 
    Un {(a,c). (a,n1) \<in>  ({(n1',n2)} Un edges b)^* \<and>  ((n2,c) \<in> ({(n1',n2)} Un edges b)^*)})" 
    by (simp add:edges_def trancl_insert)
      from a10  
      have a12:"(y,z) \<in>  (edges b)^+ Un {(a,n2). (a,n1) \<in>  (edges b)^* | (a,n1') \<in>  (edges b)^* }"   
      proof - 
    show "(y, z) \<in> ({(n1', n2)} \<union> edges b)\<^sup>+ \<union> {(a, c). (a, n1) \<in> ({(n1', n2)} \<union> edges b)\<^sup>* 
      \<and> (n2, c) \<in> ({(n1', n2)} \<union> edges b)\<^sup>*} \<Longrightarrow>
      (y, z) \<in> (edges b)\<^sup>+ \<union> {(a, n2). (a, n1) \<in> (edges b)\<^sup>* \<or> (a, n1') \<in> (edges b)\<^sup>*}"
      apply simp
      apply(erule disjE)
      apply(simp add:trancl_insert) apply blast
      apply(erule conjE) apply(drule rtranclD) 
      apply(drule rtranclD) apply(simp add:trancl_insert) 
      apply auto 
          done
      qed
      from a12 and a6 and  a3 and a4 and IH and casehyp and A  
      show ?thesis
      proof - 
    show "  \<lbrakk>(y, z) \<in> (edges b)\<^sup>+ \<union> {(a, n2). (a, n1) \<in> (edges b)\<^sup>* \<or> 
      (a, n1') \<in> (edges b)\<^sup>*}; n1' \<in> nodes b;
      strand n1 \<noteq> strand n2 \<and> n1 \<rightarrow> n2 \<and> n1 \<in> nodes b \<and>
      (\<forall>n3. n3 \<in> nodes b \<longrightarrow> (n1, n3) \<notin> edges b); n2 \<notin> nodes b;
      (y, z) \<in> (edges b)\<^sup>+ \<longrightarrow> y \<noteq> z; z = n2;
      b \<in> bundles\<rbrakk>
      \<Longrightarrow> y \<noteq> z" 
          apply simp apply(erule conjE)+
          apply(erule disjE) apply simp
          apply(auto dest:bundle_is_rtrans_closed23) done
      qed
    next
      case False
      from this have casehyp:"z\<noteq>n2" .
      from  a8 have a9:"(y,z) \<in> ({(n1, n2),(n1',n2)} \<union> edges b)\<^sup>+" 
    by(unfold edges_def, simp)
      from A and  a8 and casehyp and a4 have a10: "(y,z) \<in>  (edges b)\<^sup>+" 
      proof -  
        show "\<lbrakk>b \<in> bundles; (y, z) \<in> (edges ({n2} \<union> nodes b, {(n1, n2), (n1', n2)} \<union> edges b))\<^sup>+;
      z \<noteq> n2; n2 \<notin> nodes b\<rbrakk> \<Longrightarrow> (y, z) \<in> (edges b)^+" 
      apply -
          apply (simp add:edges_def trancl_insert)
          apply(erule disjE) apply simp apply(erule disjE, erule conjE)
          apply(drule rtranclD)+
      apply(fold edges_def) 
      apply(auto dest:bundle_is_trans_closed22) 
          apply(drule rtranclD)+    apply(simp add:trancl_insert)+
       apply(auto dest:rtranclD bundle_is_trans_closed22)
          done
      qed
      from a10 and IH show ?thesis by simp
    qed
  qed

next
  fix b n1 n2
  assume a1:"b \<in> bundles" 
    and IH:"(y, z) \<in> (edges b)\<^sup>+ \<longrightarrow> y \<noteq> z"
    and a2:" node_sign n2 = -" 
    and a3:" strand n1 \<noteq> strand n2 \<and> n1 \<rightarrow> n2 \<and> n1 \<in> nodes b \<and> (\<forall>n3. n3 \<in> nodes b \<longrightarrow> (n1, n3) \<notin> edges b)" 
    and a4:" n2 \<notin> nodes b" and a5:"index n2=0" 
  show "(y, z) \<in> (edges ({n2} \<union> nodes b, {(n1, n2)} \<union> edges b))\<^sup>+ \<longrightarrow> y \<noteq> z"
  proof
    assume a8:"(y, z) \<in> (edges ({n2} \<union> nodes b, {(n1, n2)} \<union> edges b))\<^sup>+"
    show "y\<noteq>z"
    proof(cases "z=n2")
      case True
      from this have casehyp:"z=n2" .
      from a8 have a9:"(y,z):({(n1, n2)} \<union> edges b)\<^sup>+" 
    by(unfold edges_def, simp)
      from a9  have a12:"(y,z): (edges b)^+ Un {(a,n2). (a,n1) \<in>  (edges b)^*  }"   
      proof - 
    show "(y, z) \<in> ({(n1, n2)} \<union> edges b)\<^sup>+ \<Longrightarrow> (y, z) \<in> (edges b)\<^sup>+ \<union> {(a, n2). (a, n1) \<in> (edges b)\<^sup>*}"
          apply(simp add:edges_def trancl_insert) apply blast
          done
      qed
      from a12  and  a3 and a4 and IH and casehyp and A  show ?thesis
      proof - 
    show " \<lbrakk>(y, z) \<in> (edges b)\<^sup>+ \<union> {(a, n2). (a, n1) \<in> (edges b)\<^sup>*};
      strand n1 \<noteq> strand n2 \<and> n1 \<rightarrow> n2 \<and> n1 \<in> nodes b \<and> (\<forall>n3. n3 \<in> nodes b \<longrightarrow> (n1, n3) \<notin> edges b);
      n2 \<notin> nodes b; (y, z) \<in> (edges b)\<^sup>+ \<longrightarrow> y \<noteq> z; z = n2; b \<in> bundles\<rbrakk>
      \<Longrightarrow> y \<noteq> z" 
          apply simp apply(erule conjE)+
          apply(erule disjE) apply simp
          apply(auto dest:bundle_is_rtrans_closed23) done
      qed
    next
      case False
      from this have casehyp:"z\<noteq>n2" .
      from  a8 have a9:"(y,z) \<in> ({(n1, n2)} \<union> edges b)\<^sup>+" 
    by(unfold edges_def, simp)
      from A and  a8 and casehyp and a4 have a10: "(y,z): (edges b)\<^sup>+" 
      proof -  
        show "\<lbrakk>b \<in> bundles; (y, z) \<in> (edges ({n2} \<union> nodes b, {(n1, n2)} \<union> edges b))\<^sup>+;
      z \<noteq> n2; n2 \<notin> nodes b\<rbrakk> \<Longrightarrow> (y, z) \<in> (edges b)^+"   
          apply (simp add:edges_def trancl_insert)
          apply(erule disjE) apply simp
      apply(fold edges_def)
      apply(auto dest:rtranclD bundle_is_trans_closed22)
          done
      qed
      from a10 and IH show ?thesis by simp
    qed
  qed
qed

lemma bundle_is_wf:"\<lbrakk>b \<in>  bundles \<rbrakk>\<Longrightarrow> wf  (edges b)"
  apply(rule finite_acyclic_wf)
  apply(erule bundles.induct) apply(unfold edges_def, simp+)
  apply( unfold acyclic_def,fold edges_def,auto dest:bundle_edge_is_anti2)
  done

lemma bundle_trancl_wf: "\<lbrakk>b \<in> bundles \<rbrakk>\<Longrightarrow> wf ((edges b)^+)"
  apply(blast dest:  bundle_is_wf wf_trancl)
  done



lemma bundle_minimal: "\<lbrakk>b \<in> bundles; x \<in> Q\<rbrakk>
  \<Longrightarrow>\<exists>z\<in>Q. (\<forall>y. (y\<prec>\<^sub>b z) \<longrightarrow> y \<notin> Q)"
  apply(frule bundle_trancl_wf)
  apply( simp only:wf_eq_minimal)

  done

section{*lemmas on origination assumption*}



lemma non_originate_node:"\<lbrakk>unique_originate a n; n\<noteq>n'\<rbrakk>\<Longrightarrow>
  strand n= strand n' \<and> (a \<sqsubset> node_term n' \<longrightarrow> index n < index n') |
  strand n\<noteq> strand n' \<and> non_originate a (strand n')"
  apply(unfold unique_originate_def, unfold originate_def)
  apply(case_tac "strand n=strand n'") apply(erule conjE)+ 
  apply (case_tac "(index n') <  (index n)") apply(drule_tac ?x= " n'" in spec) 
  apply(drule_tac ?x="index n'" in spec)apply(unfold strand_def index_def, simp)
  apply(case_tac "index n= index n'") 
  apply (unfold index_def,simp )
  apply(subgoal_tac  "n=n'") apply blast
  apply(simp add:node_equal strand_def index_def)
  apply simp
  apply(unfold non_originate_def  strand_def index_def)
  apply (rule disjI2)
  apply(rule conjI)
  apply simp
  apply(rule allI)
  apply(rule impI)
  apply(rule ccontr) apply(erule conjE)+
  apply(drule_tac x="na" in spec)
  apply( auto ) 
  apply(cut_tac n="node_sign (fst n', b)" in  node_sign_is_Plus_or_minus)
  apply simp
  done

lemma arith_aux_non_originate1:"\<lbrakk> (\<exists> (i::nat). P i)\<rbrakk>\<Longrightarrow> \<exists> i. (P i \<and> (\<forall> j<i. \<not> P j))"
  apply(erule_tac  exE) thm exI
  apply(rule_tac ?x=" Least P " in exI)
  apply(rule conjI)
  apply(rule LeastI) apply simp
  apply (rule allI)
  apply (auto dest: not_less_Least)
  done

lemma non_originate_strand:" non_originate a s \<Longrightarrow> 
  (\<exists> i. node_sign (s,i)=- \<and>  a \<sqsubset> node_term (s,i)  \<and> (\<forall> i0< i. \<not> a \<sqsubset> node_term  (s , i0))) |
  (\<forall> i. \<not>  a \<sqsubset> node_term (s,i))"
  apply(unfold non_originate_def)
  apply( case_tac "\<exists>i.  a \<sqsubset>  node_term (s,i) ") 
  prefer 2 apply simp
  apply(drule arith_aux_non_originate1)
  apply(erule exE)
  apply(drule_tac ?x="(s,i)" in spec) 
  apply(unfold strand_def index_def, simp add:fst_conv)
  apply(erule disjE)
  apply blast
  apply auto
  done

lemma non_originate_node1:
  "\<lbrakk>unique_originate a n; strand n\<noteq>strand n';a\<sqsubset> node_term n'\<rbrakk>\<Longrightarrow>
  \<exists> m. first_node_in_nonorigi_strand a n' m "
  apply -
  apply(subgoal_tac "n\<noteq>n'")
  prefer 2 
  apply(rule ccontr)
  apply(simp add:node_equal)
  apply(unfold first_node_in_nonorigi_strand_def)
  apply(drule_tac n'="n'" in non_originate_node,blast)
  apply(erule disjE) 
  apply simp
  apply(erule conjE)+
  apply(drule non_originate_strand)
  apply(erule disjE)
  apply(erule exE) thm exI
  apply(rule_tac x="(strand n',i)" in exI)
  apply(unfold strand_def index_def,simp)
  apply(drule_tac x="index n'" in spec)
  apply(fold strand_def)
  apply(subgoal_tac "n'=(strand n', index n')")
  apply auto
  apply(simp add:node_equal)
  apply(unfold strand_def index_def)
  apply auto
  done

  
 

lemma nonOrigNode:
  assumes 
  a1:"node_sign m2=+" and
  a2:" non_originate a (strand m2)" and
  a4:" first_node_in_nonorigi_strand a m2 m" and
  a8:"a \<sqsubset>  node_term m2"
  shows
  "index m < index m2"
proof -
from a1 a4 have b1:"index m \<noteq>index m2"
  apply -
  apply (rule ccontr)
  apply(unfold first_node_in_nonorigi_strand_def)
  apply(subgoal_tac "m=m2")
  prefer 2
  apply(simp add:node_equal)
  apply auto
  done

from a2  a4 a8  have "index m <= index m2"
  apply - 
  apply(unfold non_originate_def)
  apply(unfold first_node_in_nonorigi_strand_def)
  apply(rule ccontr)
  apply(drule_tac x="m" in spec)
  apply(erule conjE)+
  apply(drule_tac x="index m2" in spec)
  apply simp
  apply(subgoal_tac "m2=(strand m2, index m2)")
  apply(unfold strand_def index_def)
  apply auto
  done
with b1 show ?thesis
 by auto

qed
    


lemma nonOrigNode1:
  assumes 
  a2:" strand m1=strand m2" and
  a4:" first_node_in_nonorigi_strand a m2 m" and
  a8:"a \<sqsubset>  node_term m1"
  shows
  "index m <= index m1"
proof -
from a2  a4 a8 show ?thesis
  apply - 
  apply(unfold first_node_in_nonorigi_strand_def)
  apply(rule ccontr)
  apply(erule conjE)+
  apply(drule_tac x="index m1" in spec)
  apply simp
  apply(subgoal_tac "m1=(strand m2, index m1)")
  apply simp
  apply(simp add:node_equal)
  apply(unfold strand_def index_def)
  apply auto
  done
qed

section{*lemmas on the penetrator's knowledge closure property*}


lemma synth_sep_suite:
  "\<lbrakk>suite M a b; t \<in> synth M; t=MPair g h\<rbrakk>\<Longrightarrow>
  g \<in> synth M \<and> h \<in>  synth M"
  apply(case_tac "a\<sqsubset> t")
  apply(unfold suite_def component_def)
  apply(subgoal_tac "MPair g h \<notin> M")
  apply blast
  apply(rule ccontr)
  apply blast
  apply(subgoal_tac "\<not> a \<sqsubset> g \<and> \<not> a \<sqsubset> h")
  apply(subgoal_tac "g \<in> M \<and> h \<in> M")
  apply blast
  apply blast
  apply(rule ccontr, unfold parts_relation_def)
  apply( simp)
  apply(subgoal_tac "a \<in> parts {g,h}")
  apply blast
  apply(erule disjE) apply(subgoal_tac "parts {g} \<subseteq> parts {g,h}",blast) thm parts_mono
  apply( rule parts_mono, blast)
  apply(subgoal_tac "parts {h} \<subseteq> parts {g,h}",blast) thm parts_mono
  apply( rule parts_mono, blast)
  done

lemma  penetrator_strand: 
  assumes a1:" Is_penetrator_strand (strand m')" 
  and a2: "node_sign m'=+"  
  and a3: "suite M a b"
  and a5:"\<exists> m. (m\<Rightarrow>\<^sup>+m')" 
  and a6:"b \<in> bundles"
  and  a7:" \<forall>  m.(m\<Rightarrow>\<^sup>+m') \<and>node_sign m=-\<longrightarrow>(node_term m \<in>  (synth M))" 
  and  a8:" m' \<in>  nodes b"  
  
  
  shows c:"   node_term m' \<in>  (synth M)  "        
proof -
  from a8 and a6 have a8:"m' \<in> Domain" thm   bundle_node_in_Domain
    by(blast dest: bundle_node_in_Domain) 
  let ?s="strand  m'"
  from a1 have "(  Is_Tee_strand ?s | Is_Flush_strand ?s | Is_Cat_strand ?s |Is_Sep_strand ?s|
    Is_E_strand ?s | Is_D_strand ?s |Is_T_strand ?s | Is_K_strand ?s)"
    apply - apply ( simp add:penetrator_trace)  done
  moreover
  { assume a12:"Is_E_strand (strand m')"
    from a2 and a8 and  a12 have M0:"snd m'=2" 
      apply - apply( unfold Is_E_strand_def, unfold Domain_def)
      apply(unfold node_sign_def strand_def,auto)
      apply(case_tac "y=0") apply simp  
      apply (case_tac "y=1" ) apply simp apply auto   done
      from a12 and this 
      have M1:"\<exists> k h . node_term (fst m',0)=Key k \<and> node_term (fst m', 1)=h \<and> 
    node_term m'=Crypt k h" (is "\<exists> k h. ?P k h")
        by (unfold node_term_def, unfold Is_E_strand_def strand_def, auto)
      then obtain k and h where M01:" ?P k h"  by auto
      from a12  have M00:"node_sign (fst m', 0)=-\<and> node_sign (fst m', 1)=-" 
        apply - apply(unfold Is_E_strand_def,unfold  node_sign_def strand_def, auto) done
      from M0  and a8 have M000:"(fst m',1) \<in> Domain\<and>((fst m', 1) ,m'):casual2^+"
        apply- apply( rule casual2_trancl) apply simp apply simp done

       from M0  and a8 have M001:"(fst m',0) \<in> Domain\<and>((fst m', 0) ,m'):casual2^+"
        apply- apply( rule casual2_trancl) apply simp apply simp done

      from M000 and  a7  and M00 and  M01  have M3:" node_term (fst m',1)  \<in>  (synth ( M))"
    apply- apply((erule conjE)+,drule_tac x="(fst m',1)" in spec, drule mp) apply auto done

      from this and M01  have M4:"h \<in>  synth(  ( M))" by simp
      from M0  and a8 have M001:"(fst m',0) \<in> Domain\<and>((fst m', 0) ,m'):casual2^+"
        apply- apply( rule casual2_trancl) apply simp apply simp done

      from M001 and  a7  and M00 and  M01  have M3:" node_term (fst m',0)  \<in>  (synth ( M))"
    apply- apply((erule conjE)+,drule_tac x="(fst m',0)" in spec, drule mp) apply auto done
      with M01 have M6:"Key k \<in>   M" by simp
      
      
      from this and M4   have M7:"(Crypt k h) \<in> synth ( synth ( M))" by blast
      from this  and M01  have ?thesis  
    apply - apply((erule conjE)+ ,simp) apply( blast dest:synth_synthD) done
    }
    moreover
    { assume a12: "Is_D_strand (fst m')"
      from a2 and a8 and  a12 have M0:"snd m'=2" 
    apply - 
    apply( unfold Is_D_strand_def, unfold Domain_def)
    apply(unfold node_sign_def strand_def,auto)
    apply(case_tac "y=0") apply simp  
    apply (case_tac "y=1" ) apply simp 
    apply auto   done
      from a12 and this 
      have M1:"\<exists> k. \<exists> k'. \<exists> h. k'=invKey k\<and> node_term (fst m',0)= (Key k')
    \<and>node_term (fst m',1)= (Crypt k h)\<and>
    node_term  m'=h" (is "\<exists> k k'  h. ?P k k' h")
        by (unfold node_term_def, unfold Is_D_strand_def, auto)
      then obtain k and k' and h where M01:" ?P k k' h"  
    by auto
      from a12  have M00:"node_sign (fst m', 0)=-\<and> node_sign (fst m', 1)=-" 
        apply - apply(unfold Is_D_strand_def,unfold  node_sign_def, auto) done
      from M0  and a8 have M000:"(fst m',1)\<in>Domain\<and>((fst m', 1) ,m'):casual2^+"
        apply- apply( rule casual2_trancl) 
    apply simp apply simp done
      from M0  and a8 have M001:"(fst m',0)\<in>Domain\<and>((fst m', 0) ,m'):casual2^+"
        apply- apply( rule casual2_trancl) 
    apply simp apply simp done
      with a7 M00 M01 have M002:"Key (invKey k) \<in> synth M"
    apply - apply(drule_tac x="(fst m',0)" in spec)
    apply auto
    done
      then have M002:"Key (invKey k) \<in>  M"
    apply - apply blast done
    
      from M000 M00  a7 M01 have M003:"Crypt k h \<in> synth M"
    apply - apply(drule_tac x="(fst m',1)" in spec)
    apply auto
    done
      have "a \<sqsubset> Crypt k h | \<not> a \<sqsubset>  Crypt k h" by blast
      moreover
      {assume  N0:"a \<sqsubset> Crypt k h"

    from M000 and a6 and a8 and M0 
    have M51: "(strand  m',0)\<in> nodes b " 
      apply - apply( subgoal_tac "((fst m',1), m'):casual2^+") 
      apply( drule_tac i="0" in bundle_casual2_closed3 ) 
      apply assumption+
      apply simp apply blast 
      apply(blast intro:trancl_into_rtrancl) done
      
    from a3  M002 N0  M51 a1 M01
    have M6: "\<not> Crypt k h\<in>M"  
      apply -
      apply(unfold suite_def)
          apply (rule ccontr) 
      apply(erule conjE)
      apply(drule_tac x="Crypt k h" in bspec) 
      apply simp 
      apply(unfold component_def regularK_def)
      apply simp
      apply(drule_tac x="(strand m',0)" in bspec)
      apply simp
      apply(unfold Is_regular_strand_def strand_def)
      apply auto
      done

    from this and M003  and M01  have ?thesis  by blast
    }
    moreover
    {assume  N0:"\<not> a \<sqsubset> Crypt k h"
         from N0 have "\<not>  a\<sqsubset> h" 
       apply - apply(rule ccontr)
       apply (unfold parts_relation_def)
       apply auto
       done
     with a3 have "h\<in>M"
       apply - apply(unfold suite_def)
       apply blast
       done
     then have "h \<in>synth M"
       by blast
       
     with M01 have ?thesis by blast
       }
      ultimately have ?thesis by blast
    }

    moreover
    {assume a12:"Is_Cat_strand (fst m')"
      from a2 and a8 and  a12 have M0:"snd m'=2" 
    apply - apply( unfold Is_Cat_strand_def, unfold Domain_def, unfold node_sign_def,auto)
    apply(case_tac "y=0") apply simp   apply (case_tac "y=1" ) apply simp apply auto   done
      from a12 and this 
      have M1:"\<exists> g h . node_term (fst m',0)=g \<and> node_term (fst m', 1)=h 
    \<and>  node_term m'={|g,h|}" (is "\<exists> g h. ?P g h")
        by (unfold node_term_def, unfold Is_Cat_strand_def, auto)
      then obtain g and h where M01:" ?P g h"  by auto
      from a12  have M00:"node_sign (fst m', 0)=-\<and> node_sign (fst m', 1)=-" 
        apply - apply(unfold Is_Cat_strand_def,unfold  node_sign_def, auto) done

      
      from M0  and a8 have M000:"(fst m',1)\<in>Domain\<and>((fst m', 1) ,m'):casual2^+"
        apply- apply( rule casual2_trancl) apply simp apply simp done
      from M000 and  a7  and M00 and  M01 and a3  have M3:" h \<in>(synth ( M))"
    apply- apply(case_tac "a\<sqsubset> node_term (fst m', 1)")
    apply((erule conjE)+,drule_tac x="(fst m',1)" in spec, drule mp) apply auto
    done

      from M0  and a8 have M001:"(fst m',0)\<in>Domain\<and>((fst m', 0) ,m'):casual2^+"
        apply- apply( rule casual2_trancl) apply simp apply simp done
      from M001 and  a7  and M00 and  M01 and a3 have M5:"  g\<in> (synth (M))"
        apply- apply(case_tac "a\<sqsubset> node_term (fst m', 0)")
        apply((erule conjE)+,drule_tac x="(fst m',0)" in spec, drule mp) apply auto done
      
      from  M3 and  M5 and  M01  have ?thesis by (simp, blast dest:synth_synthD)
      
    }

    moreover
    {assume a12:"Is_Sep_strand (fst m')"
      from a2 and a8 and  a12 have M0:"snd m'=2|snd m'=1" 
    apply - 
    apply( unfold Is_Sep_strand_def)
    apply(unfold Domain_def, unfold node_sign_def,auto)
    apply(case_tac "y=0") apply simp   
    apply(case_tac "y=1" ) apply simp
    apply auto   done  
      from a12  
      have M00:"node_sign (fst m', 0)=-\<and>
    node_sign (fst m', 1)=+\<and> node_sign (fst m', 2)=+" 
        apply - apply(unfold Is_Sep_strand_def)
    apply(unfold  node_sign_def, auto) done
      from a12 and this 
      have M1:"\<exists> g h . node_term (fst m',0)={|g,h|} 
    \<and> node_term (fst m', 1)=g \<and>  node_term (fst m',2)=h" (is "\<exists> g h. ?P g h")
        by (unfold node_term_def, unfold Is_Sep_strand_def, auto)
      then obtain g and h where M01:" ?P g h" 
    by auto 
      from M0  have M0:"snd m'=2|snd m'=1" by simp
      moreover
      {assume a20:"snd m'=2"
    from this  and a8 have M000:"(fst m',0)\<in>Domain\<and>((fst m', 0) ,m'):casual2^+" 
      apply- apply( rule casual2_trancl) 
      apply simp apply simp done
    
    from a20  and M01 have M02: "node_term m'= h" 
      apply - apply ((erule conjE)+,simp)
      apply(drule sym) apply simp done
    
    from M000 and  a7  and M00 and  M01 and a3  have M3:" MPair g h \<in>(synth ( M))"
      apply- 
      apply((erule conjE)+,drule_tac x="(fst m',0)" in spec, drule mp) apply auto
      done

    with a3 have  "h\<in>synth M" by (blast dest:synth_sep_suite)

    
    from this and   M02  have ?thesis
      apply - 
      apply blast
      done
      }      
      moreover
      {assume a20:"snd m'=1"
    from this  and a8 have M000:"(fst m',0)\<in>Domain\<and>((fst m', 0) ,m'):casual2^+" 
      apply- apply( rule casual2_trancl)
      apply simp apply simp done
    from a20  and M01 have M02: "node_term m'= g" 
      apply - apply ((erule conjE)+,simp)
      apply(drule sym) apply simp done

    
    from M000 and  a7  and M00 and  M01 and a3  have M3:" MPair g h \<in>(synth ( M))"
      apply- 
      apply((erule conjE)+,drule_tac x="(fst m',0)" in spec, drule mp) apply auto
      done

    with a3 have  "g\<in>synth M" by (blast dest:synth_sep_suite)

    
    from this and  M02 
    have ?thesis 
      by blast
      }    
      ultimately have ?thesis by blast  
    }

    moreover
    {assume a9:"Is_Tee_strand ?s"
      from a8 and a2 and a9 
      have M0: "snd m' =2 | snd m'=1" 
    apply - apply( unfold Is_Tee_strand_def)
    apply(unfold Domain_def strand_def, auto)
    apply(case_tac "y=0") apply(unfold node_sign_def, simp)
    apply auto done
      from a9 and this 
      have "\<exists> g . node_term (fst m',0)=g \<and> node_term
    (fst m', 1)=g \<and>  node_term m'=g" (is "\<exists> g. ?P g")
        by (unfold node_term_def Is_Tee_strand_def strand_def, auto)
      then obtain g where M01:" ?P g" ..
      from a9 have M00:"node_sign (fst m', 0)=-"
    apply -
    apply(unfold Is_Tee_strand_def  node_sign_def)
    apply(unfold strand_def, auto)
    done
      have  ?thesis 
      proof - 
    from M0 have M0: "snd m' =2 | snd m'=1" by simp
    moreover 
    {assume M1:" snd m'=2" 
     from this  and a8 have M000:"(fst m',0)\<in>Domain\<and>((fst m', 0) ,m'):casual2^+" 
      apply- apply( rule casual2_trancl) 
      apply simp apply simp done
    
    from M1  and M01 have M02: "node_term m'= g" 
      apply - apply ((erule conjE)+,simp)
      done
    
    from M000 and  a7  and M00 and  M01 and a3  have M3:" g \<in>(synth ( M))"
      apply- 
      apply((erule conjE)+,drule_tac x="(fst m',0)" in spec, drule mp) apply auto
      done 
    from this and M02 have ?thesis
      by blast
    }
    moreover
    {assume M2:" snd m'=1" 
      
     from this  and a8 have M000:"(fst m',0)\<in>Domain\<and>((fst m', 0) ,m'):casual2^+" 
       apply- apply( rule casual2_trancl) 
       apply simp apply simp done
 
     from M2  and M01 have M02: "node_term m'= g" 
       apply - apply ((erule conjE)+,simp)
       done
    
     from M000 and  a7  and M00 and  M01 and a3  
     have M3:" g \<in>(synth ( M))"
       apply- 
       apply((erule conjE)+,drule_tac x="(fst m',0)" in spec, drule mp) apply auto
       done 
     from this and M02 have ?thesis
       by blast
       }
       ultimately show  ?thesis by blast
     qed
   }
   moreover
   {assume a9:"Is_T_strand ?s"
     from a2 and a9 and a8  have "snd m'=0" apply -
       apply (unfold Is_T_strand_def, unfold node_sign_def)
       apply(unfold Domain_def strand_def) apply auto done

     from  this and a5  have ?thesis
       apply - apply (erule exE ) 
       apply(drule casual2_trancl_less) 
       apply(unfold index_def,simp)
       done
   } 
   moreover
   {assume a10 :"Is_K_strand (fst m')"
     from a2 and a10 and a8  have "snd m'=0" apply -
       apply (unfold Is_K_strand_def, unfold node_sign_def)
       apply(unfold Domain_def ) 
       apply auto done

     from  this and a5  
     have ?thesis 
       apply - apply (erule exE ) 
       apply(drule casual2_trancl_less)
       apply(unfold index_def, arith)
       done
   } 
   moreover
   {assume a11:"Is_Flush_strand (fst m')"
     from a2 and a11 and a8 have ?thesis  
       apply- 
       apply(unfold Is_Flush_strand_def,unfold Domain_def, unfold node_sign_def, auto) 
       done
   }


    
   ultimately show  ?thesis by(unfold strand_def, blast)

 qed

lemma  another_penetrator_strand: 
  assumes a1:" node_term m'\<notin> (synth M)" 
  and a2: "node_sign m'=+"  
  and a3: "suite M a b"
  and a5:"\<exists> m. (m\<Rightarrow>\<^sup>+m')" 
  and a6:"b:bundles"
  and  a7:" \<forall>  m. (m\<Rightarrow>\<^sup>+m')\<and> node_sign m=-\<longrightarrow>(node_term m: (synth M))" 
  and  a8:" m'\<in> nodes b"  

  
  
  shows c:"  Is_regular_strand (strand m')    "
proof -

from a1 a2 a3 a5 a6 a7 a8  
show "  Is_regular_strand (strand m')    "      
  apply - apply(rule ccontr)
  apply(unfold Is_regular_strand_def)

  apply (blast dest: penetrator_strand)
  done
qed

section{*lemmas on path existence *}


lemma slice_arr_not_nil:" \<exists> n ns. slice_arr_cons s i len= n# ns \<and> n=(s,i)"
  apply(induct_tac len)
  apply auto
  done

lemma slic_arr_ith: "\<forall>  start i. i<=len \<longrightarrow>(slice_arr_cons s start len)!i=(s,start+i)"
  apply(induct_tac len) apply simp
  apply(rule allI)+ apply(rule impI)
  apply auto 
  apply(case_tac "i=0") apply simp  apply(drule_tac x=" Suc start" in spec) 
  apply(drule_tac x="i- 1 " in spec)  apply simp
  apply(subgoal_tac "  ((s, start) # slice_arr_cons s (Suc start) n)= ([(s, start)]@ slice_arr_cons s (Suc start) n)")
  apply(simp only:nth_append)
  apply(subgoal_tac "0<i")  apply(drule mp) apply arith  apply (simp ) apply simp apply simp done

lemma slice_arr_ith:"i<=len \<Longrightarrow>(slice_arr_cons s start len)!i=(s,start+i)"
  apply - apply(cut_tac slic_arr_ith, blast)
  done


lemma slice_arr_length: "\<forall> start. length (slice_arr_cons s start len) =len +1"
  apply(induct_tac len) apply auto
  done


lemma slice_arr_length1[simp]: " length (slice_arr_cons s start len) =len +1"
  apply(cut_tac slice_arr_length, auto)
  done

lemma slice_arr_last[simp]:
  " last (slice_arr_cons s start len)= (s,(start + len))"
 apply(subgoal_tac "(slice_arr_cons s start len) \<noteq>[]")
 prefer 2 
 apply(cut_tac slice_arr_not_nil[where  s="s" and i="start" and len="len"])
 apply(erule exE)+
 apply simp
 apply(cut_tac slice_arr_length[where s="s" and len="len"])
 apply (drule_tac x="start" in spec)
 apply(cut_tac last_conv_nth[where xs="(slice_arr_cons s start len)"])
 apply auto
 apply(cut_tac slic_arr_ith[where s="s" and len="len"])
 apply(drule_tac x="start" in spec)
 apply(drule_tac x="len" in spec)
 apply blast
done

lemma slice_arr_last2[simp]:
  "\<lbrakk>index m <= index m2;strand m=strand m2\<rbrakk>\<Longrightarrow>last ( slice_arr_cons (strand m) (index m) (index m2 - index m))=m2"
  apply(simp)
  apply(unfold strand_def index_def, simp add:node_equal)
  done

lemma slice_arr_no_empty[simp]:"slice_arr_cons s i len \<noteq>[]"
  apply(cut_tac slice_arr_not_nil[where  s="s" and i="i" and len="len"])
  apply(erule exE)+
  apply simp

  done

lemma slice_index:
  assumes a: "m\<in>set (slice_arr_cons (strand m) start i) "
  shows "index m <= start +i "
  proof -
  let ?xs="(slice_arr_cons (strand m) start i)"
  from a have b1:"\<exists> i < length ?xs. ?xs!i = m"
    apply - 
    apply(cut_tac a,simp only:in_set_conv_nth)
    done
  then obtain ia where b1:"ia< length ?xs \<and> ?xs!ia = m"
    by blast

  from b1 have "slice_arr_cons (strand m) start i ! ia =(strand m,start+ia)"
    apply -
    apply(rule slice_arr_ith)
    by simp

  with b1 have "m=(strand m,start+ia)"
    by simp

  with b1  
  show ?thesis
    apply -
    apply(simp add:node_equal)
    apply(unfold index_def)    
    apply auto
    done
qed
    

lemma slice_arr_close:
  assumes a1:" (m', ma) \<in> casual2\<^sup>+" and 
  a2:"ma \<in> set (slice_arr_cons (strand m) (index m) (index m2 - index m))" and
  a3:"index m < index m2" and
  a4:"a \<sqsubset>  node_term m'" and
  a5:"first_node_in_nonorigi_strand a m2 m"
  shows "m'\<in>set (slice_arr_cons (strand m) (index m) (index m2 - index m))"
  proof -
    let ?xs="(slice_arr_cons (strand m) (index m) (index m2 - index m))"
    
    from a2 have "(\<exists> i < length ?xs. ?xs!i = ma)"
      apply - apply( simp add:set_conv_nth,blast)
      done
    then obtain i where c1:"i < length ?xs \<and> ?xs!i = ma"
      by blast


    from a3 have c2:"length ?xs = (index m2 - index m) +1"
     apply -
     apply(cut_tac ?s="strand m" and len="index m2 -index m"  in slice_arr_length)
     apply(drule_tac x="index m" in spec)
     by simp

   with  c1 have c3:"(strand m, (index m + i))=ma"
     apply -
     apply simp
     apply(cut_tac s="(strand m)" and len="(index m2 - index m) " in slic_arr_ith)
     apply(drule_tac x="index m" in spec)
     apply(drule_tac x="i" in spec)
     apply auto
     done
thm casual2_trancl_less
   from a1 have c30:"index m' < index ma \<and> strand m' = strand ma"
     apply -     apply(blast dest:casual2_trancl_less)
     done

   with a1 c3 have c4:"strand m'=strand m"
     apply - 
     apply (simp add:node_equal)
     apply(unfold strand_def,auto)
     done

   from a5 this a4 have c5:"index m <=index m'"
     apply - apply(rule nonOrigNode1)
     apply(unfold first_node_in_nonorigi_strand_def,auto)
     done

   from c4 c30 have c6:"strand ma=strand m"
     by auto

   from a2 c6 have c7:"index ma<=index m +  (index m2 - index m)"
     apply - apply(rule_tac start="index m" and i="(index m2 - index m)" in slice_index)
     by simp
     
   from a3 c7 have c8:"index ma<=index m2"
     by simp
     

   with c5 c8   have "(\<exists> i < length ?xs. ?xs!i = m')"
     apply -
     apply(rule_tac x="index m' - index m" in exI)
     apply(rule conjI)
     apply(cut_tac ?s="strand m" and len="index m2 -index m"  in slice_arr_length)
     apply(drule_tac x="index m" in spec) 
     apply simp apply(cut_tac c30)
     apply arith
     apply(subgoal_tac "(slice_arr_cons (strand m) (index m) (index m2 - index m))!
       (index m' - index m)=(strand m, index m+ (index m' - index m))")
     prefer 2
     apply(rule slice_arr_ith) 
     apply(cut_tac c8 c30) apply arith 
     apply simp
     apply(cut_tac c4, simp add:node_equal)
     apply(unfold strand_def index_def, simp)
     done
   from this show ?thesis
     by(simp add:in_set_conv_nth)
 qed

lemma slice_arr_close1:
  assumes a1:" (m', ma) \<in> casual2\<^sup>+" and 
  a2:"ma \<in> set (slice_arr_cons (strand m) (index m) (index m2 - index m))" and
  a3:"index m <= index m2" and
  a4:"a \<sqsubset>  node_term m'" and
  a5:"unique_originate a m"
  shows "m'\<in>set (slice_arr_cons (strand m) (index m) (index m2 - index m))"
  proof -
    let ?xs="(slice_arr_cons (strand m) (index m) (index m2 - index m))"
    
    from a2 have "(\<exists> i < length ?xs. ?xs!i = ma)"
      apply - apply( simp add:set_conv_nth,blast)
      done
    then obtain i where c1:"i < length ?xs \<and> ?xs!i = ma"
      by blast


    from a3 have c2:"length ?xs = (index m2 - index m) +1"
     apply -
     apply(cut_tac ?s="strand m" and len="index m2 -index m"  in slice_arr_length)
     apply(drule_tac x="index m" in spec)
     by simp

   with  c1 have c3:"(strand m, (index m + i))=ma"
     apply -
     apply simp
     apply(cut_tac s="(strand m)" and len="(index m2 - index m) " in slic_arr_ith)
     apply(drule_tac x="index m" in spec)
     apply(drule_tac x="i" in spec)
     apply auto
     done
thm casual2_trancl_less
   from a1 have c30:"index m' < index ma \<and> strand m' = strand ma"
     apply -     apply(blast dest:casual2_trancl_less)
     done

   with a1 c3 have c4:"strand m'=strand m"
     apply - 
     apply (simp add:node_equal)
     apply(unfold strand_def,auto)
     done

   from a5 this a4 have c5:"index m <=index m'"
     apply - 
     apply(unfold unique_originate_def originate_def )
     apply(rule ccontr) apply(erule conjE)+
     apply(drule_tac x="index m'" in spec)
     apply simp
     apply(subgoal_tac "m'=(strand m, index m')")
     apply simp
     apply(simp add:node_equal)
     apply(unfold strand_def index_def, simp)
     done

   from c4 c30 have c6:"strand ma=strand m"
     by auto

   from a2 c6 have c7:"index ma<=index m +  (index m2 - index m)"
     apply - apply(rule_tac start="index m" and i="(index m2 - index m)" in slice_index)
     by simp
     
   from a3 c7 have c8:"index ma<=index m2"
     by simp
     

   with c5 c8   have "(\<exists> i < length ?xs. ?xs!i = m')"
     apply -
     apply(rule_tac x="index m' - index m" in exI)
     apply(rule conjI)
     apply(cut_tac ?s="strand m" and len="index m2 -index m"  in slice_arr_length)
     apply(drule_tac x="index m" in spec) 
     apply simp apply(cut_tac c30)
     apply arith
     apply(subgoal_tac "(slice_arr_cons (strand m) (index m) (index m2 - index m))!
       (index m' - index m)=(strand m, index m+ (index m' - index m))")
     prefer 2
     apply(rule slice_arr_ith) 
     apply(cut_tac c8 c30) apply arith 
     apply simp
     apply(cut_tac c4, simp add:node_equal)
     apply(unfold strand_def index_def, simp)
     done
   from this show ?thesis
     by(simp add:in_set_conv_nth)
 qed


lemma Complete_path_mono: "\<lbrakk>p\<in>complete_Path  a g1 \<rbrakk>\<Longrightarrow> 
  nodes g1 \<subseteq>  nodes g2\<longrightarrow> edges g1 \<subseteq> edges g2\<longrightarrow>
  p\<in>complete_Path a g2"
  apply(erule  complete_Path.induct)
  apply (blast intro: addPos1 addPos2 addnege )+
  done



lemma completePathProp:
  assumes a1:"p\<in>complete_Path a b"

  shows "unique_originate a n
         \<longrightarrow>(b\<in>bundles)
         \<longrightarrow>(\<forall> n'.((n'\<in> nodes b) \<and>last p=n' \<longrightarrow>(n'=n | (n,n'):(edges b)^+)))" (is "?cons1 p ")
  and "p \<noteq>[]"  (is "?cons2 p")
  and "\<forall>m. m\<in>(set p)\<and>node_sign m=+ \<and> a\<sqsubset> node_term m \<longrightarrow> 
          (\<forall>m'.(m'\<Rightarrow>\<^sup>+ m)\<and>a\<sqsubset>node_term  m'\<longrightarrow> m'\<in>(set p))" (is "?cons3 p ")

using a1
proof induct
  
  fix m  m2 p
  let ?p="p @ slice_arr_cons (strand m) (index m) (index m2 - index m)"
  assume 
    a2:" non_originate a (strand m2)" and
    a3:"node_sign m2 = +" and a4:" first_node_in_nonorigi_strand a m2 m" and
    a5:"p \<in> complete_Path a b" and 
    a6:"?cons1 p \<and> ?cons2 p \<and>?cons3 p "  and
    a7:"last p = m" and 
    a8:"a \<sqsubset>  node_term m2"
  show 
    " ?cons1 ?p  \<and> ?cons2 ?p \<and>?cons3 ?p "
  proof(rule conjI,(rule impI)+,rule allI)
    fix n'
    assume 
    b1:"unique_originate a n" and
    b2:"b \<in> bundles" 
    show "n' \<in> nodes b \<and> last (p @ slice_arr_cons (strand m) (index m) (index m2 -index m)) = n' \<longrightarrow>
            n' = n \<or> (n, n') \<in> (edges b)\<^sup>+"
    proof(rule impI, (erule conjE)+)
      assume
    b3:" n' \<in> nodes b" and
    b4:"last (p @ slice_arr_cons (strand m) (index m) (index m2 - index m)) = n'" 
      from a2 a3 a4 a8 have c1:"index m <= index m2"
    by(blast intro:nonOrigNode1)
      show "n' = n \<or> (n, n') \<in> (edges b)\<^sup>+"
      proof -

    from a4 have c2:"strand m2=strand m"
      by(unfold first_node_in_nonorigi_strand_def,simp)

    from c1 c2 
    have "last (p @ slice_arr_cons (strand m) (index m) (index m2 - index m))=m2"
      apply - 
      apply(subgoal_tac "slice_arr_cons (strand m) (index m) (index m2 - index m)\<noteq>[]")
      prefer 2 apply simp
      apply(simp add:last_appendR) 
      apply(simp add:node_equal)
      apply(unfold  index_def strand_def, auto)
      done

    with b4 have c4:"n'=m2" by simp
      thm  bundle_casual2_closed3
    with b2 c1 c2 b3 have c3:"m:nodes b\<and>(m,m2):(edges b Int casual2)^*" 
      apply - 
      apply(subgoal_tac "(strand m2, index m)\<in>nodes b \<and> ((strand m2, index m),m2)\<in>(edges b Int casual2)^*")
      prefer 2
      apply (blast dest: bundle_casual2_closed3)
      apply(simp add:node_equal)
      apply(unfold strand_def index_def,auto)
      done

    with b1 b2 a6 a7 have "m=n | (n,m): (edges b)^+"
      apply - apply blast done

    with c3 have "m2=n | (n,m2): (edges b)^+"
      apply - apply(subgoal_tac "(m, m2) \<in> (edges b)^*")
      apply(drule rtranclD) apply(erule disjE) 
      apply(erule disjE) apply simp apply simp
      apply(erule disjE) apply simp 
      apply(blast dest:transitive_closure_trans)
      apply(subgoal_tac "(edges b \<inter> casual2)\<^sup>* \<subseteq>  (edges b)\<^sup>*")
      apply(blast ) thm rtrancl_mono
      apply(blast intro!: rtrancl_mono)
      done

    with c4 show ?thesis by simp
      qed
    qed
  next
    show "?cons2 ?p \<and>?cons3 ?p "
    proof
      show "?cons2 ?p" by simp
    next 
      show "?cons3 ?p"
      proof((rule allI,rule impI)+,(erule conjE)+)
    fix ma m'       
    assume c1:"(m', ma) \<in> casual2\<^sup>+" and
      c2:"a \<sqsubset>  node_term m'" and
      c3:"ma\<in>set ?p" and
      c4:" node_sign ma = +" and 
      c5:" a \<sqsubset>  node_term ma"
    show "m' \<in> set ?p"
    proof -
      from c3 
      have "ma:set p | ma\<in>set (slice_arr_cons (strand m) (index m) (index m2 - index m))"
        by simp
      moreover
      {
        assume d1:"ma:set p "
        from a6   d1 c4 c5 c1 c2
        have "m':set p"
          apply -
          apply(erule conjE)+
          apply(drule_tac x="ma" in spec)
          apply blast
          done
        then  have ?thesis 
          by simp
      }
      moreover
      {
        assume d1:"ma:set (slice_arr_cons (strand m) (index m) (index m2 - index m))"
        from a2 a3 a4 a8 have d2:"index m < index m2"
          by(blast intro:nonOrigNode)
        from c1 d1 c2 a4 d2
        have "m':set  (slice_arr_cons (strand m) (index m) (index m2 - index m))"
          apply - thm  slice_arr_close
          apply(rule_tac m'="m'" and ma="ma" and ?m2.0="m2" and m="m" in slice_arr_close)
          apply auto
          done
        then have ?thesis by simp
      }
      ultimately show ?thesis
        by blast
    qed
      qed
    qed
  qed
next
  fix m  m2 
  let ?p="slice_arr_cons (strand m) (index m) (index m2 - index m)"
  assume 
    a2:"unique_originate a m" and
    a3:"node_sign m2 = +" and a4:"strand m2 = strand m" and
    a8:"a \<sqsubset>  node_term m2" 
  show   " ?cons1 ?p  \<and> ?cons2 ?p \<and>?cons3 ?p "
  proof(rule conjI,(rule impI)+,rule allI)
    fix n'
    assume 
    b1:"unique_originate a n" and
    b2:"b \<in> bundles" 
    show "n' \<in> nodes b \<and> last ( slice_arr_cons (strand m) (index m) (index m2 -index m)) = n' \<longrightarrow>
            n' = n \<or> (n, n') \<in> (edges b)\<^sup>+"
    proof(rule impI, (erule conjE)+)
      assume
    b3:" n' \<in> nodes b" and
    b4:"last ( slice_arr_cons (strand m) (index m) (index m2 - index m)) = n'" 

      show "n' = n \<or> (n, n') \<in> (edges b)\<^sup>+"
      proof -
    from a3 a2 a4 a8  have c1:"index m <= index m2"
          apply -
      apply(unfold unique_originate_def originate_def)
      apply(rule ccontr) 
      apply(erule conjE)+ 
      apply(drule_tac x="index m2" in spec)
      apply simp apply(subgoal_tac "strand m=strand m2")
      apply(thin_tac "strand m2 = strand m") apply simp
      apply(subgoal_tac "m2=(strand m2,index m2)")
          apply (simp add:node_equal) 
      prefer 2 apply simp
      apply(unfold strand_def index_def,simp)
      done
    from c1 a4 
    have "last ( slice_arr_cons (strand m) (index m) (index m2 - index m))=m2"
      apply -  
      apply (simp only:node_equal) 
      apply(unfold strand_def index_def,auto)
      done
    with b4 have c4:"n'=m2" by simp

    with b2 c1 a4 b3 have c3:"m\<in>nodes b\<and>(m,m2):(edges b Int casual2)^*" 
      apply - 
      apply(subgoal_tac "(strand m2, index m):nodes b \<and> ((strand m2, index m),m2):(edges b Int casual2)^*")
      prefer 2
      apply (blast dest: bundle_casual2_closed3)
      apply(rule conjI)
      apply(subgoal_tac "m=(strand m,index m)")
      apply(simp add:node_equal)      
      apply(unfold strand_def index_def,simp)
      apply(simp add:node_equal)
      done

    from b1 a2 have "m=n" 
      apply -
      apply(unfold unique_originate_def,blast )
      done

    with c3 c4  have "m=n | (n,m): (edges b)^+"
      apply - apply blast done

    with c3 have "m2=n | (n,m2): (edges b)^+"
      apply - apply(subgoal_tac "(m, m2) \<in> (edges b)^*")
      apply(drule rtranclD) apply(erule disjE) 
      apply(erule disjE) apply simp apply simp
      apply(erule disjE) apply simp 
      apply(blast dest:transitive_closure_trans)
      apply(subgoal_tac "(edges b \<inter> casual2)\<^sup>* \<subseteq>  (edges b)\<^sup>*")
      apply(blast ) thm rtrancl_mono
      apply(blast intro!: rtrancl_mono)
      done

    with c4 show ?thesis by simp
      qed
    qed
  next
    show "?cons2 ?p \<and>?cons3 ?p "
    proof
      show "?cons2 ?p" by simp
    next 
      show "?cons3 ?p"
      proof((rule allI,rule impI)+,(erule conjE)+)
    fix ma m'       
    assume c1:"(m', ma) \<in> casual2\<^sup>+" and
      c2:"a \<sqsubset>  node_term m'" and
      c3:"ma:set ?p" and
      c4:" node_sign ma = +" and 
      c5:" a \<sqsubset>  node_term ma"
    show "m' \<in> set ?p"
    proof -
      from a2  a4 a8 have d2:"index m <= index m2"
        apply - apply(unfold unique_originate_def originate_def )
        apply(rule ccontr) apply(erule conjE)+
        apply(drule_tac x="index m2" in spec)
        apply simp
        apply(subgoal_tac "m2=(strand m, index m2)")
        apply simp
        apply(simp add:node_equal)
        apply(unfold strand_def index_def, auto)
        done
      (*from  a4 have "index m\<noteq>index m2"
        apply -
        apply(rule ccontr)
        apply(simp add:node_equal)
        done
      with d2 have "index m < index m2"
        by simp*)
      from c1 c2 c3 a2 a4 this thm  slice_arr_close1
      show ?thesis
        apply - thm slice_arr_close1
        apply(rule_tac m'="m'" and ma="ma" and ?m2.0="m2" and m="m" in slice_arr_close1)
          apply auto
          done
      qed   
    qed
      qed
    qed
next
  fix m1  m2 p
  
  

  assume 
    a2:"m1\<rightarrow> m2" and a3:"m1:nodes b" and a4:"(m1,m2)\<in>edges b" and
    a5:"p \<in> complete_Path a b" and 
    a6:"?cons1 p  \<and> ?cons2 p \<and> ?cons3 p" and
    a7:"last p = m1" and 
    a8:"a \<sqsubset>  node_term m2" 
  let ?p = "p @ [m2]"
  show 
    " ?cons1 ?p  \<and> ?cons2 ?p \<and>?cons3 ?p "
  proof(rule conjI,(rule impI)+,rule allI)
    fix n'
    assume 
    b1:"unique_originate a n" and
    b2:"b \<in> bundles" 
    show "n' \<in> nodes b \<and> last (p @ [m2]) = n' \<longrightarrow>
            n' = n \<or> (n, n') \<in> (edges b)\<^sup>+"
    proof(rule impI, (erule conjE)+)
      assume
    b3:" n' \<in> nodes b" and
    b4:"last (p @ [m2] ) = n'" 

      show "n' = n \<or> (n, n') \<in> (edges b)\<^sup>+"
      proof -     
    from a6  have "last (p @ [m2] ) =m2" 
      by simp
    with b4 have c0:"m2=n'" by simp
    from b3 b4 have c1: "m2\<in>nodes b" by simp
    from a3 a6 a7 b1 b2
    have "m1 = n \<or> (n, m1) \<in> (edges b)\<^sup>+"
      by  blast thm  bundle_edge_is_unique
    
    with  a4 have "(n,m2):(edges b)^+"
      by (blast dest:transitive_closure_trans)
    with c0 show ?thesis by simp  
      qed
    qed
  next
    show "?cons2 ?p \<and>?cons3 ?p "
    proof
      show "?cons2 ?p" by simp
    next 
      show "?cons3 ?p"
      proof((rule allI,rule impI)+,(erule conjE)+)
    fix ma m'       
    assume c1:"(m', ma) \<in> casual2\<^sup>+" and
      c2:"a \<sqsubset>  node_term m'" and
      c3:"ma\<in>set ?p" and
      c4:" node_sign ma = +" and 
      c5:" a \<sqsubset>  node_term ma"
    show "m' \<in> set ?p"
    proof -
      from c3 
      have "ma:set p | ma=m2"
        by auto
      moreover
      {
        assume d1:"ma\<in>set p "
        from a6   d1 c4 c5 c1 c2
        have "m':set p"
          apply -
          apply(erule conjE)+
          apply(drule_tac x="ma" in spec)
          apply blast
          done
        then  have ?thesis 
          by simp
      }
      moreover
      {
        assume d1:"ma=m2"
        from a2 have "node_sign m2=-"
          by (blast dest:recv_node_is_minus)
        from this d1 c4  have ?thesis by simp
      }
      ultimately show ?thesis
        by blast
    qed
      qed
    qed
  qed
qed


    
lemma path_exist:
  assumes A:"b\<in>bundles" 
  shows C: "unique_originate a n\<longrightarrow>
  (\<forall> n'.( n'\<in> nodes b\<and> a \<sqsubset> node_term n'\<longrightarrow>
           (\<exists> p.  p\<in> (complete_Path a b)  \<and> (last p)=n'))) "
  using A
proof induct
  case Nil show ?case by(unfold nodes_def, auto)
next
  fix b n1 n2
  assume a1: "b \<in> bundles"    and 
    a2: " unique_originate a n \<longrightarrow>
    (\<forall>n'. n' \<in> nodes b \<and> a \<sqsubset>  node_term n' \<longrightarrow>
    (\<exists>p. p\<in> (complete_Path a b)  \<and> (last p)=n') ) " and
    a3: "node_sign n2 = +" and a4:" n2 \<in> Domain" 
    and a5:" n2 \<notin> nodes b" and a6:" 0 <index n2" 
    and a7:" n1 \<in> nodes b"  and a8:" n1 \<Rightarrow> n2"
  let ?b'="({n2} \<union> nodes b, {(n1, n2)} \<union> edges b)"
  show "unique_originate a n \<longrightarrow> 
    (\<forall>n'. n' \<in> nodes ?b' \<and> a \<sqsubset>  node_term n' \<longrightarrow>
          (\<exists>p. p \<in> complete_Path a ?b' \<and> last p = n'))"
  proof(rule impI,rule allI,rule impI,erule conjE)
    fix n'
    assume 
      b1:" unique_originate a n" and 
      b2:"n' \<in> nodes ?b'" and
      b3:"a \<sqsubset>  node_term n'"    
    show "\<exists>p. p \<in> complete_Path a ?b' \<and> last p = n'"
    proof -
      
      from b2 have "n'\<in>nodes b |n'=n2" 
	by(unfold nodes_def,auto)
     
      moreover
      {
	assume c1:"n'\<in>nodes b "
	from c1 b1 b2 b3 a2
	have "(\<exists>p. p\<in> (complete_Path a b)  \<and> (last p)=n') " (is "\<exists> p. ?P p b")
	  by blast
	then obtain p where c2:"?P p b"
	  by blast
	then have c3:"?P p ?b'" thm Complete_path_mono
	  apply - apply(subgoal_tac "(nodes b ) \<subseteq> nodes  ?b' ")
	  apply(subgoal_tac "(edges b ) \<subseteq> edges  ?b'")
	  apply(erule conjE)
	  apply (blast dest:Complete_path_mono) 
	  apply(unfold nodes_def edges_def,auto)
	  done
	then have ?thesis by blast
      }
      moreover
      {
	assume c1:"n'=n2"
	have "strand n2=strand n | strand n2\<noteq>strand n" by simp
	moreover
	{ assume d1:"strand n2=strand n"
	  have "n2=n|n2\<noteq>n" by simp
	  moreover
	  {
            assume e1:"n2=n"
	    have e2: "(slice_arr_cons (strand n) (index n) (index n - index n))\<in>complete_Path a ?b'"
	      apply - 
              apply(rule addPos2)
              apply(assumption)+
              prefer 2 apply(cut_tac b1,unfold unique_originate_def originate_def,simp)
              prefer 2 apply(cut_tac b1,unfold unique_originate_def originate_def,simp)
              apply(cut_tac c1 b3,simp)
              done
            have "(slice_arr_cons (strand n) (index n) (index n - index n))=[n]"
              apply (simp add:node_equal)
	      apply(unfold strand_def index_def,auto)
              done
            with c1 e2 e1 have ?thesis 
              apply - 
              apply(rule_tac x="[n]" in exI)
              by simp
	  }
	  moreover
	  {
            assume e2:"n2\<noteq>n"       
            from e2 b1 c1 b3 d1
            have e0:"index n<=index n2 " 
              apply - thm  non_originate_node
              apply(drule_tac ?n'="n2" in  non_originate_node)
              apply blast 
              apply(unfold non_originate_def)
              
              apply auto
              done
            have "(slice_arr_cons (strand n) (index n) (index n2 - index n))\<in>complete_Path a ?b'"
              apply - 
              apply(rule addPos2)
              apply(assumption)+
              prefer 2 apply assumption 
              apply(cut_tac c1 b3,simp)
              done
        
            with e0 c1  d1 have ?thesis 
              apply -  
              apply(rule_tac x="slice_arr_cons (strand n) (index n) (index n2 - index n)" in exI)
              apply simp
              apply(simp add:node_equal)
              apply(unfold strand_def index_def, auto)
              done
	  }
	  ultimately have ?thesis by blast
	}
	moreover
	{
	  assume d2:" strand n2\<noteq>strand n"
	  from c1 b3 have d3:"a \<sqsubset> node_term n2"
            by simp
        
	  from d2 b1 this 
	  have d4:"\<exists> m. first_node_in_nonorigi_strand a n2 m" thm non_originate_node1
            apply -
            apply(rule non_originate_node1)
            by auto

	  then obtain m where 
            d4:"first_node_in_nonorigi_strand a n2 m"
            by blast

        
	  from d2 have "n2\<noteq>n"
            apply - 
            apply(rule ccontr)
            apply(simp add:node_equal)
            done

	  from this d2 b1 d3  
	  have d5:" non_originate  a (strand n2)"
	    apply -
	    apply(drule non_originate_node)
	    apply auto
	    done  
       
	  from d5 d3
	  have "index m < index n2"
	    apply - 
	    apply(rule_tac a="a" in  nonOrigNode)
	    apply assumption+
	    done

	  with a3   d4
	  have d01:"index m < index n2"
	    apply -
	    apply(subgoal_tac "index m \<noteq>index n2")
	    apply arith
	    apply(rule ccontr)
	    apply(unfold first_node_in_nonorigi_strand_def)
	    apply(subgoal_tac "m=n2")
	    apply(simp add:node_equal)+
	    done
	  from this and a8 
	  have "index m <=index n1"
	    apply - 
	    apply(unfold casual2_def,simp)
	    done

	  with a7 a1 a8
	  have "m\<in>nodes b \<and> a\<sqsubset> node_term m"
	    apply -
	    apply(subgoal_tac "(strand n1, index m)\<in>nodes b")
	    apply(cut_tac d4)
	    apply(unfold first_node_in_nonorigi_strand_def)
	    apply(unfold casual2_def)
	    apply(subgoal_tac "(strand n1, index m)=m")
	    apply simp
	    apply(simp add:node_equal)
	    apply(unfold strand_def index_def, simp)
	    apply(fold strand_def index_def)
	    apply(blast dest:bundle_casual2_closed3)
	    done
	  from a2 this b1
	  have " (\<exists>p. p: (complete_Path a b)  \<and> (last p)=m)  " (is "\<exists> p. ?P p b m")
	    by blast
     
	  then obtain p where d50 :"?P p b m"
	    by blast
	  then have d6:"?P p ?b' m" thm Complete_path_mono
	    apply - apply(subgoal_tac "(nodes b ) \<subseteq> nodes  ?b' ")
	    apply(subgoal_tac "(edges b ) \<subseteq> edges  ?b'")
	    apply(erule conjE)
	    apply (blast dest:Complete_path_mono) 
	    apply(unfold nodes_def edges_def,auto)
      done

     let ?p'="p@(slice_arr_cons (strand m) (index m) (index n2 - index m))"

     from d5 a3 d50 d6 b3 c1 d4 have "?P ?p' ?b' n2"
       apply -
       apply(rule conjI)
       apply(rule addPos1)
       prefer 7
       apply(subgoal_tac "slice_arr_cons (strand m) (index m) (index n2 - index m)\<noteq>[]")
       apply( simp add:node_equal)
       apply(cut_tac d01 d4 )
       apply(unfold strand_def index_def first_node_in_nonorigi_strand_def)  
       apply auto
       done

     with c1 have ?thesis by blast
       }
       ultimately have ?thesis by blast
     }
     ultimately show ?thesis by blast
   qed
 qed
next
  fix b n2
  assume a1: "b \<in> bundles"    and 
    a2: " unique_originate a n \<longrightarrow>
    (\<forall>n'. n' \<in> nodes b \<and> a \<sqsubset>  node_term n' \<longrightarrow>
    (\<exists>p. p\<in> (complete_Path a b)  \<and> (last p)=n') ) " and
    a3: "node_sign n2 = +"  
    and a5:" n2 \<notin> nodes b" 
    and a7:" n2 \<in>Domain"  and a8:"index n2 = 0"
  show " unique_originate a n \<longrightarrow>
    (\<forall>n'. n' \<in> nodes ({n2} \<union> nodes b, edges b) \<and> a \<sqsubset>  node_term n' \<longrightarrow>
    (\<exists>p. p \<in> complete_Path a ({n2} \<union> nodes b, edges b) \<and> last p = n'))"
    proof(rule impI,rule allI,rule impI,erule conjE)
    fix n'
    assume 
      b1:" unique_originate a n" and 
      b2:"n' \<in> nodes ({n2} \<union> nodes b,  edges b)" and
      b3:"a \<sqsubset>  node_term n'"    
    show "\<exists>p. p \<in> complete_Path a ({n2} \<union> nodes b, edges b) \<and> last p = n'"
    proof -
      let ?b'="({n2} \<union> nodes b,  edges b)"
      from b2 have "n':nodes b |n'=n2" 
    by(unfold nodes_def,auto)
     
      moreover
      {
    assume c1:"n'\<in>nodes b "
    from c1 b1 b2 b3 a2
    have "(\<exists>p. p\<in> (complete_Path a b)  \<and> (last p)=n') " (is "\<exists> p. ?P p b")
      by blast
    then obtain p where c2:"?P p b"
      by blast
    then have c3:"?P p ?b'" thm Complete_path_mono
      apply - apply(subgoal_tac "(nodes b ) \<subseteq> nodes  ?b' ")
      apply(subgoal_tac "(edges b ) \<subseteq> edges  ?b'")
      apply(erule conjE)
      apply (blast dest:Complete_path_mono) 
      apply(unfold nodes_def edges_def,auto)
      done
    then have ?thesis by blast
      }
      moreover
      {
    assume c1:"n'=n2"
    from a3 b1 c1 b3 a8
    have c2:"n=n2"
      apply - 
      apply(subgoal_tac "originate a n2")
      apply(unfold  unique_originate_def,blast)
      apply(unfold originate_def)
      apply auto
      done
     have e2: "(slice_arr_cons (strand n) (index n) (index n - index n))\<in>complete_Path a ?b'"
	      apply - 
              apply(rule addPos2)
              apply(assumption)+
              prefer 2 apply(cut_tac b1,unfold unique_originate_def originate_def,simp)
              prefer 2 apply(cut_tac b1,unfold unique_originate_def originate_def,simp)
              apply(cut_tac c1 b3,simp)
              done
            have "(slice_arr_cons (strand n) (index n) (index n - index n))=[n]"
              apply (simp add:node_equal)
	      apply(unfold strand_def index_def,auto)
              done
    
    with c1 c2 e2 have ?thesis 
      apply - 
      apply(rule_tac x="[n]" in exI)
      by simp
    }
    ultimately show ?thesis by blast
      qed
    qed
  next
    fix b n1 n1' n2
    assume a1: "b \<in> bundles"    and 
      a2: " unique_originate a n \<longrightarrow>
      (\<forall>n'. n' \<in> nodes b \<and> a \<sqsubset>  node_term n' \<longrightarrow>
      (\<exists>p. p\<in> (complete_Path a b)  \<and> (last p)=n') ) " and
      a3: "node_sign n2 = - " and
      a4:"strand n1 \<noteq> strand n2 \<and> n1 \<rightarrow> n2 \<and> n1 \<in> nodes b \<and>
      (\<forall>n3. n3 \<in> nodes b \<longrightarrow> (n1, n3) \<notin> edges b)"
      and a5:" n2 \<notin> nodes b" and a6:" 0 <index n2" 
      and a7:" n1' \<in> nodes b"  and a8:" n1' \<Rightarrow> n2"
    show "unique_originate a n \<longrightarrow> 
      (\<forall>n'. n' \<in> nodes ({n2} \<union> nodes b, {(n1, n2), (n1', n2)} \<union> edges b) \<and> a \<sqsubset>  node_term n' \<longrightarrow>
      (\<exists>p. p \<in> complete_Path a ({n2} \<union> nodes b, {(n1, n2), (n1', n2)} \<union> edges b) \<and> last p = n'))"
    proof(rule impI,rule allI,rule impI,erule conjE)
      fix n'
      assume 
    b1:" unique_originate a n" and 
    b2:"n' \<in> nodes ({n2} \<union> nodes b, {(n1, n2), (n1', n2)} \<union> edges b)" and
    b3:"a \<sqsubset>  node_term n'"    
      show "\<exists>p. p \<in> complete_Path a ({n2} \<union> nodes b, {(n1, n2), (n1', n2)} \<union> edges b) \<and> last p = n'"
      proof -
    let ?b'="({n2} \<union> nodes b, {(n1, n2), (n1', n2)} \<union> edges b)"
    from b2 have "n':nodes b |n'=n2" 
      by(unfold nodes_def,auto)
     
    moreover
    {
      assume c1:"n'\<in>nodes b "
      from c1 b1 b2 b3 a2
      have "(\<exists>p. p\<in> (complete_Path a b)  \<and> (last p)=n') " (is "\<exists> p. ?P p b")
        by blast
      then obtain p where c2:"?P p b"
        by blast
      then have c3:"?P p ?b'" thm Complete_path_mono
        apply - apply(subgoal_tac "(nodes b ) \<subseteq> nodes  ?b' ")
        apply(subgoal_tac "(edges b ) \<subseteq> edges  ?b'")
        apply(erule conjE)
        apply (blast dest:Complete_path_mono) 
        apply(unfold nodes_def edges_def,auto)
        done
      then have ?thesis by blast
    }
    moreover
    {
      assume c1:"n'=n2"
      from b3 a4 c1 have c2:"a\<sqsubset> node_term n1"
        apply - 
        apply( unfold casual1_def,auto)
        done
      with  b1 a3 b3 a4 a2
      have "(\<exists>p. p: (complete_Path a b)  \<and> (last p)=n1) " (is "\<exists> p. ?P p b n1 ")
        by blast
      then obtain p where c3:"?P p b n1"
        by blast
      then have c4:"?P p ?b' n1" thm Complete_path_mono
        apply - apply(subgoal_tac "(nodes b ) \<subseteq> nodes  ?b' ")
        apply(subgoal_tac "(edges b ) \<subseteq> edges  ?b'")
        apply(erule conjE)
        apply (blast dest:Complete_path_mono) 
        apply(unfold nodes_def edges_def,auto)
        done
      have "?P (p@[n2]) ?b' n2"
        apply -
        apply(rule conjI)
        apply(rule_tac  ?m1.0="n1" in addnege)
        apply(cut_tac a4)
        apply simp
        apply(cut_tac a4)
        apply(unfold nodes_def edges_def)
        apply simp 
        apply simp
        apply(cut_tac c4)
        apply(unfold nodes_def edges_def,simp)
        apply(cut_tac c1 b3,simp)
        apply(cut_tac c4,simp)
        apply simp
        done
      with c1 have ?thesis by blast
    }
    ultimately show ?thesis
      by blast
      qed
    qed
  next
    fix b n1 n2
    assume a1: "b \<in> bundles"    and 
      a2: " unique_originate a n \<longrightarrow>
      (\<forall>n'. n' \<in> nodes b \<and> a \<sqsubset>  node_term n' \<longrightarrow>
      (\<exists>p. p\<in> (complete_Path a b)  \<and> (last p)=n') ) " and
      a3: "node_sign n2 = - " and
      a4:"strand n1 \<noteq> strand n2 \<and> n1 \<rightarrow> n2 \<and> n1 \<in> nodes b \<and>
      (\<forall>n3. n3 \<in> nodes b \<longrightarrow> (n1, n3) \<notin> edges b)"
      and a5:" n2 \<notin> nodes b" and a6:" index n2=0" 
    show "unique_originate a n \<longrightarrow> 
      (\<forall>n'. n' \<in> nodes ({n2} \<union> nodes b, {(n1, n2)} \<union> edges b) \<and> a \<sqsubset>  node_term n' \<longrightarrow>
      (\<exists>p. p \<in> complete_Path a ({n2} \<union> nodes b, {(n1, n2)} \<union> edges b) \<and> last p = n'))"
    proof(rule impI,rule allI,rule impI,erule conjE)
      fix n'
      assume 
    b1:" unique_originate a n" and 
    b2:"n' \<in> nodes ({n2} \<union> nodes b, {(n1, n2)} \<union> edges b)" and
    b3:"a \<sqsubset>  node_term n'"    
      show "\<exists>p. p \<in> complete_Path a ({n2} \<union> nodes b, {(n1, n2)} \<union> edges b) \<and> last p = n'"
      proof -
    let ?b'="({n2} \<union> nodes b, {(n1, n2)} \<union> edges b)"
    from b2 have "n'\<in>nodes b |n'=n2" 
      by(unfold nodes_def,auto)
     
    moreover
    {
      assume c1:"n'\<in>nodes b "
      from c1 b1 b2 b3 a2
      have "(\<exists>p. p\<in> (complete_Path a b)  \<and> (last p)=n') " (is "\<exists> p. ?P p b")
        by blast
      then obtain p where c2:"?P p b"
        by blast
      then have c3:"?P p ?b'" thm Complete_path_mono
        apply - apply(subgoal_tac "(nodes b ) \<subseteq> nodes  ?b' ")
        apply(subgoal_tac "(edges b ) \<subseteq> edges  ?b'")
        apply(erule conjE)
        apply (blast dest:Complete_path_mono) 
        apply(unfold nodes_def edges_def,auto)
        done
      then have ?thesis by blast
    }
    moreover
    {
      assume c1:"n'=n2"
      from b3 a4 c1 have c2:"a\<sqsubset> node_term n1"
        apply - 
        apply( unfold casual1_def,auto)
        done
      with  b1 a3 b3 a4 a2
      have "(\<exists>p. p\<in> (complete_Path a b)  \<and> (last p)=n1) " (is "\<exists> p. ?P p b n1 ")
        by blast
      then obtain p where c3:"?P p b n1"
        by blast
      then have c4:"?P p ?b' n1" thm Complete_path_mono
        apply - apply(subgoal_tac "(nodes b ) \<subseteq> nodes  ?b' ")
        apply(subgoal_tac "(edges b ) \<subseteq> edges  ?b'")
        apply(erule conjE)
        apply (blast dest:Complete_path_mono) 
        apply(unfold nodes_def edges_def,auto)
        done
      have "?P (p@[n2]) ?b' n2"
        apply -
        apply(rule conjI)
        apply(rule_tac  ?m1.0="n1" in addnege)
        apply(cut_tac a4)
        apply simp
        apply(cut_tac a4)
        apply(unfold nodes_def edges_def)
        apply simp 
        apply simp
        apply(cut_tac c4)
        apply(unfold nodes_def edges_def,simp)
        apply(cut_tac c1 b3,simp)
        apply(cut_tac c4,simp)
        apply simp
        done
      with c1 have ?thesis by blast
    }
    ultimately show ?thesis
      by blast
      qed
    qed     
  qed

section{*main lemmas on gerneral authentication tests*}
lemma general_auth_test: 
  assumes a1:"b\<in>bundles" and  a2:" unique_originate a n" 
  and a3:"suite M a b" 
  and a4:"n'\<in>nodes b"
  and a5:"a\<sqsubset>node_term n'"
  and a6:"  node_term n'\<notin> (synth M)"
  and a8:"node_term n :(synth M)"

  shows
  "\<exists>  m' m.( n \<preceq>\<^sub>b m  \<and>  m' \<preceq>\<^sub>b n' \<and> m \<Rightarrow>\<^sup>+ m' \<and> node_sign m'=+ 
              \<and>( strand m'\<noteq>strand n\<longrightarrow>  node_sign m=- )
              \<and> a \<sqsubset> node_term m \<and> a \<sqsubset> node_term m'
              \<and>    node_term m'\<notin> (synth M) \<and>  node_term m\<in> (synth M)
              \<and>Is_regular_strand (strand m')
              \<and>(\<forall> y.(y \<prec>\<^sub>b m') \<longrightarrow> \<not> a\<sqsubset> node_term y|(node_term y)\<in>(synth M)))
  "
  (*\<not> a\<sqsubset> node_term y|*)
 
proof -

  let ?P="{ x. (x \<preceq>\<^sub>b n')  \<and> a \<sqsubset> node_term x\<and>    node_term x\<notin> (synth M)  }"

  from a4 a5 a6  have M0:"n'\<in> ?P" by blast
 
  from a1 and this 
  have M1:"\<exists> z. z \<in> ?P \<and>  (\<forall> y.(y \<prec>\<^sub>b z) \<longrightarrow> y\<notin>?P)"  (is "\<exists>z. z \<in> ?P \<and>  ?Q z")
    apply - apply (drule_tac x="n'" and  Q="?P" in bundle_minimal) 
    apply simp apply blast done

  from this obtain m' where M2: "m'\<in>?P \<and>?Q m'" 
    by auto

  from a1 a4 M2 have M2a:"m'\<in>nodes b"
    apply - apply(simp,(erule conjE)+)
    apply(blast dest:bundle_is_rtrans_closed23)
    done

  from this have M3:" node_sign m'= +"
  proof(rule_tac P="node_sign m' = +" in  ccontr)
    assume a00:"node_sign m' \<noteq> +" 
    from a00 have M4:"node_sign m'=-" 
      apply - apply (cut_tac node_sign_is_Plus_or_minus[where n="node_sign m'"], blast) done

    
    from a1 and M2 and M4 and M2a
    have M5:"\<exists> m''. m''\<in>nodes b \<and> m'' \<rightarrow>  m'\<and>(m'', m') \<in> edges b" 
      apply - apply ((erule conjE)+,simp only:mem_Collect_eq) 
      apply(erule conjE) thm bundle_edge_is_exist
      apply( drule_tac b="b" and ?n2.0="m'" in bundle_edge_is_exist) 
      apply blast done 
    from this obtain m'' 
      where M6:"m''\<in>nodes b \<and> m'' \<rightarrow>  m'\<and>(m'', m') \<in> edges b" 
      by auto
    from this 
    have M7:"node_term m'=node_term m''" 
      by (unfold casual1_def,simp)
    from M6 M2 have "(m'',n'):(edges b)^*"
      apply - apply(blast intro:transitive_closure_trans)
      done
    from M7 and M2 and M6 and this  have " m''\<in>?P" 
      apply - apply(simp  )
      done
    from this and M2 and M6 show  "False"
      apply - thm r_into_trancl 
      apply(erule conjE)+ 
      apply(drule_tac x="m''" in spec) 
      apply (blast dest:r_into_trancl)
      done
  qed
  from M2  have  M4:"a\<sqsubset>  node_term m'" by simp

  

  

  have M8a:"\<forall> m. ((m,m'):casual2^+\<and>node_sign m=- \<longrightarrow> node_term m\<in>synth M)"(is "\<forall> m. ?P m")
    proof(rule allI,rule impI)
      fix ma
      assume N1:"(ma, m') \<in> casual2\<^sup>+\<and>node_sign ma=- "
      show "node_term ma\<in>synth M"
    proof -
      from N1 a1 M2a
      have N2:" ma\<in> nodes b \<and>(ma,m'):(edges b)^+" thm bundle_casual2_trancl
        apply - apply(rule bundle_casual2_trancl)
        apply auto
        done
      with M2 have N2a:"(ma,n'): (edges b)^*" 
        apply - apply simp
        apply(rule transitive_closure_trans)
        apply auto
        done
      have "a\<sqsubset> node_term ma | \<not> a \<sqsubset>node_term ma " by blast
      moreover
      {
        assume N3:"a\<sqsubset> node_term ma"
        from N2 M2 N2a N3 have ?thesis
          apply - apply(erule conjE)+
          apply(drule_tac x="ma" in spec)
          apply auto
          done
      }
      moreover
      {
        assume N3:"\<not> a\<sqsubset> node_term ma"
        from N3 a3 have "node_term ma \<in> M"
          apply -
          apply(unfold suite_def,auto)
          done
        then have ?thesis
          by blast
      }
      ultimately show ?thesis by blast
    qed
      qed
          
      

  

  from M2 a8 have "m'\<noteq>n"
    apply - apply(rule ccontr,auto)
    done

  
  from this and a2
  have "strand n= strand m' \<and> (a \<sqsubset> node_term m' \<longrightarrow> index n < index m') |
     strand n\<noteq> strand m' \<and> non_originate a (strand m')"
    apply - apply(rule  non_originate_node)
    by auto

  moreover
  {assume M9:" strand n\<noteq> strand m' \<and> non_originate a (strand m')"
    thm non_originate_node1
    have "\<exists>m. first_node_in_nonorigi_strand a m' m" 
      apply - apply(cut_tac a2 M9 M4)
      apply(blast intro!:non_originate_node1)
      done

    then obtain m where M10:"first_node_in_nonorigi_strand a m' m"
      by blast

    then have M11:"a \<sqsubset> node_term m"
      by(unfold first_node_in_nonorigi_strand_def,auto)

    from M10 M4 have M12:"index m <=index m'"
      apply - apply(rule nonOrigNode1)
      apply auto
      done
    thm bundle_casual2_closed3
    have "(strand m', index m) \<in> nodes b \<and> 
      ((strand m', index m), m') \<in> (edges b \<inter> casual2)\<^sup>*"
      apply - apply(cut_tac a1 M12 M2a)
      apply(rule  bundle_casual2_closed3)
      by auto

    with M10 have M13:"m\<in>nodes b\<and>(m,m'):(edges b \<inter> casual2)\<^sup>*"
      apply - 
      apply(subgoal_tac "(strand m', index m)=m",simp)
      apply(unfold first_node_in_nonorigi_strand_def)
      apply(simp add:node_equal)
      apply(unfold strand_def index_def,auto)
      done

   then  have M14:"(m,m'):( casual2)\<^sup>*" 
      apply - thm  rtrancl_mono 
      apply(subgoal_tac "(edges b \<inter> casual2)^*\<subseteq> casual2^*")
      apply (blast intro!: rtrancl_mono)+done

    
    have M15: "m\<noteq>m'"
      apply - apply(rule ccontr)
      apply(cut_tac M3 M10)
      apply(unfold first_node_in_nonorigi_strand_def, auto)
      done

    with M14 have M16:"(m,m'):( casual2)\<^sup>+" 
      apply - apply(blast dest:rtranclD )
      done

    from M13 have M17:"(m,m'):(edges b)\<^sup>*" 
      apply - thm  rtrancl_mono 
      apply(subgoal_tac "(edges b \<inter> casual2)^*\<subseteq> (edges b)^*")
      apply (blast intro!: rtrancl_mono)+done

    
    with M15 have M18:"(m,m'):( edges b)\<^sup>+" 
      apply - apply(blast dest:rtranclD )
      done
    
    from M17  M2 M14 have M19:"(m,n'):(edges b)^*"
      by(blast dest:transitive_closure_trans)

    from M19 M18 M11 M2
    have M20:"node_term m\<in>synth M"
      apply -
      apply(erule conjE)
      apply(drule_tac x="m" in spec)
      by blast

    from M16 have M21:"\<exists> m.(m,m'):casual2^+" 
      by blast 
    
    have M22:"Is_regular_strand (strand m')"
      apply - thm another_penetrator_strand apply(rule another_penetrator_strand)
      apply(cut_tac M2) apply blast
      apply(cut_tac M3) apply simp
      apply assumption+
      apply(cut_tac M21, assumption+)
      apply(cut_tac M8a, assumption+)
      apply(cut_tac M2a, auto)
      done

    
    from M11  and a1 and a2 and M13
    have M23: "(\<exists> p.  p\<in> (complete_Path a b)  \<and> (last p)=m)" (is "\<exists> p. ?R p")
      thm path_exist
      by(blast dest:path_exist)

    from this obtain p where "?R p" by auto
    from a1 this a2 M13 have M24: "n=m|(n,m):(edges b)^+" thm completePathProp
      apply - apply((erule conjE), drule completePathProp)
      apply blast
      done

    from this have M24:"(n,m):(edges b)^*"  by (blast dest:trancl_into_rtrancl)
    
    have ?thesis 
      apply - 
      apply(rule_tac x="m'" in exI) 
      apply(rule_tac x="m" in exI) 
      apply(cut_tac M22 M20 M19 M16 M2 M3 M10 M24 M10) 
      apply(unfold first_node_in_nonorigi_strand_def)
      apply(rule conjI, blast)+  
      apply(rule allI,rule impI)
      apply(subgoal_tac "(y,n'):(edges b)^*") 
      apply blast
      apply(subgoal_tac " (y, n') \<in> (edges b)\<^sup>+")
      apply simp thm transitive_closure_trans 
      apply simp
      apply(rule_tac a="y" and b="m'" and c="n'" in  trancl_rtrancl_trancl)
      apply auto
      done
  }
  moreover
  { assume
    M9:"strand n= strand m' \<and> (a \<sqsubset> node_term m' \<longrightarrow> index n < index m') "
    from M4 M9
    have M10:"index n < index m'"
      by simp

    from this  have M12:"index n <=index m'"
      by simp
    thm bundle_casual2_closed3
    have "(strand m', index n) \<in> nodes b \<and> 
      ((strand m', index n), m') \<in> (edges b \<inter> casual2)\<^sup>*"
      apply - apply(cut_tac a1 M12 M2a)
      apply(rule  bundle_casual2_closed3)
      by auto

    with M9 have M13:"n:nodes b\<and>(n,m'):(edges b \<inter> casual2)\<^sup>*"
      apply - 
      apply(subgoal_tac "(strand m', index n)=n",simp)
      
      apply(simp add:node_equal)
      apply(unfold strand_def index_def,auto)
      done

   then  have M14:"(n,m'):( casual2)\<^sup>*" 
      apply - thm  rtrancl_mono 
      apply(subgoal_tac "(edges b \<inter> casual2)^*\<subseteq> casual2^*")
      apply (blast intro!: rtrancl_mono)+done

    
    have M15: "n\<noteq>m'"
      apply - apply(rule ccontr)
      apply(cut_tac M3 M10)
      apply( auto)
      done

    with M14 have M16:"(n,m'):( casual2)\<^sup>+" 
      apply - apply(blast dest:rtranclD )
      done

    then have M17:"\<exists> m.(m,m'):( casual2)\<^sup>+" 
      by blast
    
    have M22:"Is_regular_strand (strand m')"
      apply - thm another_penetrator_strand apply(rule another_penetrator_strand)
      apply(cut_tac M2) apply blast
      apply(cut_tac M3) apply simp
      apply assumption+
      apply(cut_tac M17, assumption+)
      apply(cut_tac M8a, assumption+)
      apply(cut_tac M2a, auto)
      done

    
    have ?thesis 
      apply - 
      apply(rule_tac x="m'" in exI) 
      apply(rule_tac x="n" in exI) 

      apply(cut_tac M22  M16 M2 M3 M9 M10 a2 a8) 
      apply(unfold unique_originate_def originate_def)
      apply(rule conjI,simp)+
      apply(rule allI,rule impI)
      apply(subgoal_tac "(y,n'):(edges b)^*") 
      apply blast 
      apply(subgoal_tac " (y, n') \<in> (edges b)\<^sup>+")
      apply simp thm transitive_closure_trans 
      apply simp
      apply(rule_tac a="y" and b="m'" and c="n'" in  trancl_rtrancl_trancl)
      apply auto
      done
  }
  ultimately show ?thesis by blast
      
qed


 


lemma general_auth_test1: 
  assumes a1:"b\<in> bundles" and  a2:" unique_originate a n" 
  and a3:"suite M a b" 
  and a4:"n'\<in> nodes b"
  and a5:"a\<sqsubset>node_term n'"
  and a6:"  node_term n'\<notin>  (synth M)"
  and a8:"node_term n \<in>(synth M)"
  and a9:"\<forall>  n0. strand n0=strand n\<and> node_sign n0=+\<and>
  a \<sqsubset> node_term n0 \<and> n0\<in>  nodes b\<longrightarrow> (node_term n0\<in>  (synth M))"

  shows
  "\<exists>  m' m.((n,m):(edges b)^* \<and> (m',n'):(edges b)^*\<and>(m, m'):casual2^+  \<and> node_sign m'=+ 
              \<and>( strand m'\<noteq>strand n \<and>  node_sign m=- )
              \<and> a \<sqsubset> node_term m \<and> a \<sqsubset> node_term m'
              \<and>    node_term m'\<notin>  (synth M) \<and>  node_term m\<in>  (synth M)
              \<and>Is_regular_strand (strand m')
              \<and>(\<forall> y.(y,m'):(edges b)^+ \<longrightarrow>\<not>  a\<sqsubset> node_term y| (node_term y)\<in> (synth M)))"
apply -
apply(cut_tac a1 a2 a3 a4 a5 a6  a8)
apply(drule general_auth_test)
apply assumption+
apply(erule exE)+
apply(rule_tac x="m'" in exI)
apply(rule_tac x="m" in exI)
apply(cut_tac a9)
apply(case_tac "strand m'=strand n")
apply(erule conjE)+
apply(drule_tac x="m'" in spec)
apply(subgoal_tac "m'\<in> nodes b")
apply blast
apply(cut_tac a1 a4)
apply(blast dest:bundle_is_rtrans_closed23)
by (blast )

lemma general_auth_test2: 
  assumes a1:"b\<in>bundles" and  a2:" unique_originate a n" 
  and a3:"suite M a b" 
  and a4:"n'\<in>nodes b"
  and a5:"a\<sqsubset>node_term n'"
  and a6:"  node_term n'\<notin> (synth M)"
  and a8:"node_term n :(synth M)"

  shows
  "\<exists>  m' m.( n \<preceq>\<^sub>b m  \<and>  m' \<preceq>\<^sub>b n' \<and> m \<Rightarrow>\<^sup>+ m' \<and> node_sign m'=+ 
              \<and>( strand m'\<noteq>strand n\<longrightarrow>  node_sign m=- )
              \<and> a \<sqsubset> node_term m \<and> a \<sqsubset> node_term m'
              \<and>    node_term m'\<notin> (synth M) \<and>  node_term m\<in> (synth M)
              \<and>Is_regular_strand (strand m')
              \<and>(\<forall> y.(y \<prec>\<^sub>b m') \<longrightarrow> (node_term y)\<in>(synth M)))
  "
proof -
  have "\<exists>  m' m.( n \<preceq>\<^sub>b m  \<and>  m' \<preceq>\<^sub>b n' \<and> m \<Rightarrow>\<^sup>+ m' \<and> node_sign m'=+ 
              \<and>( strand m'\<noteq>strand n\<longrightarrow>  node_sign m=- )
              \<and> a \<sqsubset> node_term m \<and> a \<sqsubset> node_term m'
              \<and>    node_term m'\<notin> (synth M) \<and>  node_term m\<in> (synth M)
              \<and>Is_regular_strand (strand m')
              \<and>(\<forall> y.(y \<prec>\<^sub>b m') \<longrightarrow> \<not> a\<sqsubset> node_term y|(node_term y)\<in>(synth M)))
  " (is "\<exists>  m' m. ?P m' m")
  proof(rule  general_auth_test)qed
  then obtain m' and m where b1:"?P m' m" 
    by blast
  from this show ?thesis
  proof(rule_tac x="m'" in exI,rule_tac x="m" in exI,(rule_tac conjI,simp)+)
    show "\<forall>y. y \<prec>\<^sub>b m' \<longrightarrow> node_term y \<in> synth M"
    proof(rule allI,rule impI)
      fix y
      assume c1:" y \<prec>\<^sub>b m'"
      from b1 c1 have "\<not> a\<sqsubset> node_term y|(node_term y)\<in>(synth M)"
	by blast
      then show "node_term y \<in> synth M"
	apply -
	apply(erule disjE)
	apply(subgoal_tac "node_term y \<in> M")
	apply blast
	apply(cut_tac a3, unfold suite_def,auto)
	done
    qed
  qed
qed

lemma synthM:
  assumes 
  a8:"m \<in>(synth M)" and
  a9:"a=Nonce N \<or> a=Key k'" and
  a11:"M= {Crypt k h} \<union> {m. \<not> a \<sqsubset> m}  "
  shows " a \<sqsubset>m \<longrightarrow> Crypt k h  \<sqsubset>m" (is "?P m ")
  using a8
  proof induct
    fix X
    assume b1:"X \<in> M "
    show "?P X"
      apply(cut_tac b1 a11,unfold parts_relation_def,auto)
      done
  next
    fix agt
    show "?P (Agent agt)"
      apply(cut_tac  a9  a11,unfold parts_relation_def,auto)
      done
  next
    fix n
    show "?P (Number n)"
      apply(cut_tac a9  a11,unfold parts_relation_def,auto)
      done
  next
    fix X
    show "?P (Hash X)"
      apply(cut_tac a9  a11,unfold parts_relation_def,auto)
      done
  next
    fix X Y  
     assume c1:"a \<sqsubset>  X \<longrightarrow> Crypt k h \<sqsubset>  X" and 
       c2:"  a \<sqsubset>  Y \<longrightarrow> Crypt k h \<sqsubset>  Y"
     show "?P \<lbrace>X, Y\<rbrace>"
     proof
       assume d1:" a \<sqsubset>  \<lbrace>X, Y\<rbrace>"
       have "a \<sqsubset>  X | a \<sqsubset> Y "
       apply(cut_tac a9 d1 ,unfold parts_relation_def,simp)
       apply(erule disjE)
       apply simp
       apply(subgoal_tac "parts {X, Y}=parts{X} \<union> parts{Y}")
       apply simp
       apply(force simp add:parts_insert2)
       apply(subgoal_tac "parts {X, Y}=parts{X} \<union> parts{Y}")
       apply simp
       apply(force simp add:parts_insert2)
       done
       with c1 c2 show " Crypt k h \<sqsubset>  \<lbrace>X, Y\<rbrace>"
	 apply -
	 apply(erule disjE, unfold parts_relation_def)
	 apply(subgoal_tac "parts {X} \<subseteq> parts {X,Y}",force,simp add:parts_mono )
	 apply(subgoal_tac "parts {Y} \<subseteq> parts {X,Y}",force,simp add:parts_mono )
	 done
     qed
   next
     fix K X
     assume c1:"a \<sqsubset>  X \<longrightarrow> Crypt k h \<sqsubset>  X"
     show "?P (Crypt K X)"
     proof
       assume d1:" a \<sqsubset>  Crypt K X"
       have "a \<sqsubset>  X "
	 apply(cut_tac a9  d1 ,unfold parts_relation_def,auto)
	 done
       with c1  show " Crypt k h \<sqsubset> (Crypt K X)"
	 by(unfold parts_relation_def,auto)
     qed
   qed
    
lemma synthM1:
  assumes 
  a8:"m \<in>(synth M)" and
  a9:"a=Nonce N \<or> a=Key k'" and
  a11:"M= {Crypt k h} \<union> {m. \<not> a \<sqsubset> m}  " and
  a12:" a \<sqsubset>m "
  shows " Crypt k h  \<sqsubset>m" (is "?P m ")
  apply -
  apply(cut_tac a8 a9 a11 a12,blast dest: synthM)
  done


lemma out_auth_tast:
  assumes a1:"b\<in> bundles" and  a2:" unique_originate a n" 
  and a3:"regularK (invKey k) b" 
  
  and a4:"n'\<in> nodes b"
  and a5:"a\<sqsubset>node_term n'"
  and a6:"\<not> Crypt k h \<sqsubset>  node_term n'"
  and a7:"node_sign n'= -"
  and a8:"node_term n \<in>(synth M)"
  and a9:"a=Nonce N \<or> a=Key k'" 
  and a10:"M= {Crypt k h} \<union> {m. \<not> a \<sqsubset> m}  "
  and a11:"a\<sqsubset>h" 
  and a12:"n\<Rightarrow>n'"
  and a14:"node_sign n'=-"
 
  shows
  "\<exists>  m' m.( n \<preceq>\<^sub>b m  \<and>  m' \<preceq>\<^sub>b n' \<and> m \<Rightarrow>\<^sup>+ m' \<and> node_sign m'=+ 
              \<and>(   node_sign m=- )
              \<and> a \<sqsubset> node_term m \<and> a \<sqsubset> node_term m'
              \<and>    node_term m'\<notin> (synth M) \<and>  node_term m\<in> (synth M)
              \<and>Is_regular_strand (strand m')
              \<and>(\<forall> y.(y \<prec>\<^sub>b m') \<longrightarrow>(node_term y)\<in>(synth M)))
  "
proof -
  have b1:"suite M a b"
    apply(cut_tac a10 a3,unfold suite_def component_def)
    by auto
  
  have b2:" node_term n' \<notin>  synth M"
  proof(rule ccontr)
    assume c1:"\<not> node_term n' \<notin> synth M "
    from c1 have c2:"node_term n'\<in> synth M "
      by simp
    have "Crypt k h \<sqsubset>  node_term n'"
    proof(rule synthM1,cut_tac c2)qed
    with a6 show False by simp
  qed
  have "\<exists>  m' m.( n \<preceq>\<^sub>b m  \<and>  m' \<preceq>\<^sub>b n' \<and> m \<Rightarrow>\<^sup>+ m' \<and> node_sign m'=+ 
    \<and>( strand m'\<noteq>strand n\<longrightarrow>  node_sign m=- )
    \<and> a \<sqsubset> node_term m \<and> a \<sqsubset> node_term m'
    \<and>    node_term m'\<notin> (synth M) \<and>  node_term m\<in> (synth M)
    \<and>Is_regular_strand (strand m')
    \<and>(\<forall> y.(y \<prec>\<^sub>b m') \<longrightarrow> (node_term y)\<in>(synth M)))
    " (is "\<exists>  m' m. ?P m' m")
  proof(cut_tac b1 b2, rule  general_auth_test2 )qed  
  then obtain m' and m where b3:"?P m' m" by blast
  have "strand m'\<noteq> strand n"
  proof(rule ccontr)
    assume c1:"\<not> strand m'\<noteq> strand n"
    from c1 have c2:"strand m'=strand n"
      by simp
    have c2a:"m' \<noteq> n'  "
      apply(rule ccontr)
      apply(cut_tac b3 a14,simp)
      done
    have c2b:"m' \<preceq>\<^sub>b n' "
      apply(cut_tac b3 ,simp)
      done

    with c2a have c3:"m' \<prec>\<^sub>b n' "
      apply -  thm rtrancl_eq_or_trancl
      apply(simp add: rtrancl_eq_or_trancl)
      done
    
    have c4:"index m' < index n'"
    proof(rule ccontr)
      assume d1:" \<not> index m' < index n'"
      with a1  a4  c2b have d2:"m' \<in> nodes b"  
	by (blast dest:bundle_is_rtrans_closed23)
      from d1 have d3:"index n' <=index m'" by simp thm bundle_casual2_closed3
      with d2 a1 have "((strand m', index n'),m') \<in> (edges b \<inter> casual2)\<^sup>*"
	thm  bundle_casual2_closed3
      proof(drule_tac bundle_casual2_closed3,auto)qed
      then have e1:"((strand m', index n'),m') \<in> (edges b)\<^sup>*"
	apply -
	apply(subgoal_tac " (edges b \<inter> casual2)\<^sup>* \<subseteq>  (edges b)\<^sup>*")
	apply blast
	apply(rule rtrancl_mono)
	by blast
      have e2:"(strand m', index n')=n'"
	apply(simp add:node_equal)
	apply(cut_tac c2 a12, unfold casual2_def strand_def index_def,auto)
	done
      with e1 have "n' \<preceq>\<^sub>b m'"
	by simp
      with c3 have "n' \<prec>\<^sub>b n'"
	by (blast intro:transitive_closure_trans)
      with a1 show False
	by (blast dest:bundle_edge_is_anti2)
    qed
  
    have c5:"index n'=Suc (index n)" 
      by(cut_tac a12,unfold casual2_def, simp)
    with c4 have c6:"index m' <= index n"
      by arith
    have c7:"m'=(strand n, index m')"
      apply(cut_tac c1,simp add:node_equal)
      apply(unfold strand_def index_def, auto)
      done
    have c8:"index n <=index m'"
      apply(rule ccontr)
      apply(cut_tac b3 a2 c1 c2, unfold unique_originate_def originate_def)
      apply(subgoal_tac "\<not> a \<sqsubset>node_term (strand n, index m')")
      prefer 2
      apply simp
      apply(cut_tac c7)     
      by (simp )
    with c6 have "index n =index m'" 
      by simp
    with c2  have "n=m'" by (simp add:node_equal)
    with a8 b3 show False
      by simp
  qed
  with b3 show ?thesis
    apply -
    apply(rule_tac x="m'" in exI)
    apply(rule_tac x="m" in exI)
    by blast
qed


lemma unsolicited_test: 
  assumes a1:"b\<in> bundles" and
  a2:"n\<in>  nodes b" and 
  a3:"Crypt k h \<sqsubset> node_term n" and 
  a4:" regularK k b" 
  shows c:"\<exists> n'. (n',n):(edges b)^* \<and>n'\<in>  (nodes b) \<and> 
  Crypt k h \<sqsubset> node_term n'\<and> 
  Is_regular_strand (strand n')
  \<and> node_sign n'=+ \<and>
  (\<forall> y.(y,n'):(edges b)^+ \<longrightarrow>\<not> Crypt k h \<sqsubset> node_term y)"
proof -
  let ?t="Crypt k h"
  let ?P="{ x. (x,n):(edges b)^* \<and>x\<in> nodes b  \<and> ?t \<sqsubset> node_term x  }"
  from a1 and a2 and a3  
  have M1:"\<exists> z. z \<in> ?P \<and>  (\<forall> y.(y,z)\<in> (edges b)^+ \<longrightarrow>y\<notin>?P)"  (is "\<exists> z. z \<in> ?P \<and>  ?Q z")
    apply - 
    apply (drule_tac x="n" and  Q="?P" in bundle_minimal) 
    apply simp 
    apply blast 
    done
  from this obtain m' where M2: "m'\<in> ?P \<and>?Q m'"
    by auto
  from this have M3:" node_sign m'= +"
  proof(rule_tac P="node_sign m' = +" in  ccontr)
    assume a00:"node_sign m' \<noteq> +" 
    from a00 have M4:"node_sign m'=-" 
      apply - 
      apply (cut_tac node_sign_is_Plus_or_minus[where n="node_sign m'"], blast)
      done
    from a1 and M2 and M4 
    have M5:"\<exists> m''. m''\<in>nodes b \<and> m'' \<rightarrow>  m'\<and>(m'', m') \<in> edges b" 
      apply - 
      apply ((erule conjE)+,simp only:mem_Collect_eq) 
      apply(erule conjE) thm bundle_edge_is_exist  
      apply( drule_tac b="b" and ?n2.0="m'" in bundle_edge_is_exist)
      apply(subgoal_tac "m'\<in> nodes b")
      apply blast  
      apply(cut_tac a1 a2)  thm bundle_is_rtrans_closed23
      apply(blast dest:  bundle_is_rtrans_closed23)
      done 
    from this obtain m'' 
      where M6: "m''\<in>nodes b \<and> m'' \<rightarrow>  m'\<and>(m'', m') \<in> edges b"
      by auto
    from this have M7:"node_term m'=node_term m''" 
      by (unfold casual1_def,auto)
    from this and M2 and M6   
    have " m''\<in>?P" 
      apply - apply(simp only:mem_Collect_eq) 
      apply(blast dest:transitive_closure_trans )
      done
    from this and M2 and M6 show  "False" 
      apply - thm r_into_trancl 
      apply(erule conjE)+  
      apply(drule_tac x="m''" in spec) 
      apply (blast dest:r_into_trancl)done
  qed
  
  from M2 and a1 thm penetrator_strand
  have a8:"m'\<in>Domain"  thm bundle_node_in_Domain
    by(blast dest: bundle_node_in_Domain)
 
  from this and a4  
  have M03:"Is_regular_strand (strand m')" thm ccontr
  proof(rule_tac ?P="Is_regular_strand (strand m')" in ccontr)
    assume a6:"\<not> Is_regular_strand (strand m') "
    show False  
    proof -
      let ?s="strand m'"
      from a6 have ab1:"Is_penetrator_strand ?s" 
    by(unfold Is_regular_strand_def, simp)
      from ab1 have "(  Is_Tee_strand ?s | Is_Flush_strand ?s | Is_Cat_strand ?s |Is_Sep_strand ?s|
        Is_E_strand ?s | Is_D_strand ?s |Is_T_strand ?s | Is_K_strand ?s)"
    by (unfold penetrator_trace,simp)
      moreover
      { assume a12:"Is_E_strand (strand  m')"
        from M3 and a8 and  a12 have M0:"snd m'=2" 
          apply - apply( unfold Is_E_strand_def, unfold Domain_def)
          apply( unfold node_sign_def strand_def index_def,auto)
          apply(case_tac "y=0") 
      apply simp   
      apply (case_tac "y=1" ) 
      apply simp 
      apply auto   done
        from a12 and this
    have N1:"\<exists> k h . node_term (fst m',0)=Key k 
      \<and> node_term (fst m', 1)=h \<and>  node_term m'=Crypt k h" (is "\<exists> k h. ?P k h")
          by (unfold node_term_def, unfold Is_E_strand_def 
        strand_def index_def, auto)
        then obtain k0 and h0 where N01:" ?P k0 h0"  by auto

        
        from a12  have N00:"node_sign (fst m', 0)=-\<and> node_sign (fst m', 1)=-" 
          apply -
      apply(unfold Is_E_strand_def,
        unfold  node_sign_def strand_def index_def, auto) done

        from M0  and a8 have N000:"(fst m',1)\<in>Domain\<and>((fst m', 1) ,m'):casual2^+"
          apply- 
      apply( rule casual2_trancl) 
      apply simp
      apply simp done

        from a1 and M0 and M2 
    have N2:"(fst m',0)\<in> fst b\<and> ((fst m',0),m'): (edges b)^+" 
          apply - 
      apply (simp only:mem_Collect_eq,(erule conjE)+,
        fold strand_def index_def nodes_def,
        drule_tac i="0" and ?n2.0="m'" in bundle_casual2_closed3)
          apply( simp, arith) 
      apply( subgoal_tac "((strand m', 0), m') \<in> (edges b )^*" )
      prefer 2 
      apply(blast intro:rtrancl_mono1) 
      apply(thin_tac "(m', n) \<in> (edges b)\<^sup>*")
          apply(drule_tac R="edges b" in  rtranclD)
      apply (erule disjE)   
      apply( simp add:node_equal)
      apply(unfold strand_def index_def)
          apply auto
      done 

    with M2 have N2a:"((fst m',0),n): (edges b)^*"
      apply -
      apply(subgoal_tac "((fst m',0),n): (edges b)^+")
      prefer 2
      apply(blast intro:transitive_closure_trans )
      apply(blast dest:trancl_into_rtrancl)
      done
      

        from a1 and M0 and M2 
    have N3:"(fst m',1)\<in> fst b \<and> ((fst m',1),m'): (edges b)^+" 
          apply - 
      apply (simp only:mem_Collect_eq,(erule conjE)+,
         fold strand_def index_def nodes_def,
        drule_tac i="1" and ?n2.0="m'" in bundle_casual2_closed3)
      apply( simp, arith) 
      apply( subgoal_tac "((strand m', 1), m') \<in> (edges b )^*" )
      prefer 2 
      apply(blast intro:rtrancl_mono1) 
      apply(thin_tac "(m', n) \<in> (edges b)\<^sup>*")
          apply(drule_tac R="edges b" in  rtranclD)
      apply (erule disjE)   
      apply( simp add:node_equal)
      apply(unfold strand_def index_def)
          apply auto
      done 

    with M2 have N3a:"((fst m',1),n): (edges b)^*"
      apply -
      apply(subgoal_tac "((fst m',1),n): (edges b)^+")
      prefer 2
      apply(blast intro:transitive_closure_trans )
      apply(blast dest:trancl_into_rtrancl)
      done
          
    
        from N01 and M2 have "?t\<sqsubset> h0 | ?t=Crypt k0 h0"
          apply-  apply simp  
      apply(erule conjE)+ 
      apply(drule  body_subterm) 
      apply( auto) 
      done

        moreover 
        {assume a13:" ?t \<sqsubset> h0"
          
          from N01 and a13 have N4:" ?t \<sqsubset> node_term (fst m', 1)" by simp

          from this and M2  and N3 N3a have  False 
        by (unfold nodes_def,blast         )

          from this have ?thesis by blast
        }
        moreover 
        {assume a13:" ?t=Crypt k0 h0"
          
          from N01 and a13 have N4:"node_term ( fst m', 0)=Key k" by simp

          from this and a4  and ab1 and N2 N2a
      have  False 
        apply - 
        apply(unfold regularK_def Is_regular_strand_def 
          nodes_def strand_def)
        apply (drule_tac x="(fst m',0)" in bspec,blast )
        apply auto
        done

          from this 
      have ?thesis
        by blast
        }
        ultimately have ?thesis ..
      }
      moreover
      { assume a12:"Is_D_strand (strand m')"
    from M3 and a8 and  a12 have M0:"snd m'=2" 
          apply - 
      apply( unfold Is_D_strand_def, unfold Domain_def, 
        unfold node_sign_def strand_def index_def,auto)
          apply(case_tac "y=0") 
      apply simp   
      apply (case_tac "y=1" ) 
      apply simp 
      apply auto 
      done
    from a12 and this
    have M1:"\<exists> k. \<exists> k'. \<exists> h. k'=invKey k\<and> 
      node_term (fst m',0)= (Key k')\<and>node_term (fst m',1)= (Crypt k h)\<and>
          node_term  m'=h" (is "\<exists> k k'  h. ?P k k' h")
          by (unfold node_term_def,
        unfold Is_D_strand_def strand_def index_def, auto)
    then obtain k0 and k0' and h0
      where M01:" ?P k0 k0' h0"  
      by auto
    from a12  
    have M00:"node_sign (fst m', 0)=-\<and> node_sign (fst m', 1)=-" 
          apply - 
      apply(unfold Is_D_strand_def,unfold  node_sign_def
        strand_def index_def, auto) 
      done
    from M0  and a8 
    have M000:"(fst m',1)\<in>Domain\<and>((fst m', 1) ,m'):casual2^+"
          apply- 
      apply( rule casual2_trancl)
      apply simp 
      apply simp 
      done
    from a1 and M0 and M2 
    have N3:"(fst m',1)\<in> fst b \<and> ((fst m',1),m'): (edges b)^+" 
          apply - 
      apply (simp only:mem_Collect_eq,(erule conjE)+,
         fold strand_def index_def nodes_def,
        drule_tac i="1" and ?n2.0="m'" in bundle_casual2_closed3)
      apply( simp, arith) 
      apply( subgoal_tac "((strand m', 1), m') \<in> (edges b )^*" )
      prefer 2 
      apply(blast intro:rtrancl_mono1) 
      apply(thin_tac "(m', n) \<in> (edges b)\<^sup>*")
          apply(drule_tac R="edges b" in  rtranclD)
      apply (erule disjE)   
      apply( simp add:node_equal)
      apply(unfold strand_def index_def)
          apply auto
      done 

    with M2 have N3a:"((fst m',1),n): (edges b)^*"
      apply -
      apply(subgoal_tac "((fst m',1),n): (edges b)^+")
      prefer 2
      apply(blast intro:transitive_closure_trans )
      apply(blast dest:trancl_into_rtrancl)
      done
          
        from M01 and M2 have "?t \<sqsubset> h0" by simp

        from this and M01 have "?t \<sqsubset> node_term (fst m',1)" 
      by (unfold parts_relation_def,simp)

        from this and M2  and N3 N3a 
    have  False
      apply - 
      apply(unfold regularK_def Is_regular_strand_def 
        nodes_def strand_def)
      apply(erule conjE)+
      apply(drule_tac x="(fst m',1)" in spec )
      apply auto
      done       

        from this have ?thesis by blast
      }
      moreover
      {assume a12:"Is_Cat_strand (strand  m')"
        from M3 and a8 and  a12 have M0:"snd m'=2" 
          apply - 
      apply( unfold Is_Cat_strand_def, unfold Domain_def,
        unfold node_sign_def strand_def index_def,auto)
          apply(case_tac "y=0") 
      apply simp   
      apply (case_tac "y=1" )
      apply simp apply auto   done

        from a12 and this
    have M1:"\<exists> g h . node_term (fst m',0)=g \<and>
      node_term (fst m', 1)=h \<and>
      node_term m'={|g,h|}" (is "\<exists> g h. ?P g h")
          by (unfold node_term_def, 
        unfold Is_Cat_strand_def strand_def index_def, 
        auto)
        then obtain g0 and h0
      where M01:" ?P g0 h0"  
      by auto

        from M0  and a8 
    have N000:"(fst m',1)\<in>Domain\<and>((fst m', 1) ,m'):casual2^+"
          apply- apply( rule casual2_trancl)
      apply simp 
      apply simp done


        from a1 and M0 and M2 
    have N2:"(fst m',0)\<in> fst b\<and> ((fst m',0),m'): (edges b)^+" 
          apply - 
      apply (simp only:mem_Collect_eq,(erule conjE)+,
        fold strand_def index_def nodes_def,
        drule_tac i="0" and ?n2.0="m'" in bundle_casual2_closed3)
          apply( simp, arith) 
      apply( subgoal_tac "((strand m', 0), m') \<in> (edges b )^*" )
      prefer 2 
      apply(blast intro:rtrancl_mono1) 
      apply(thin_tac "(m', n) \<in> (edges b)\<^sup>*")
          apply(drule_tac R="edges b" in  rtranclD)
      apply (erule disjE)   
      apply( simp add:node_equal)
      apply(unfold strand_def index_def)
          apply auto
      done 

    with M2 have N2a:"((fst m',0),n): (edges b)^*"
      apply -
      apply(subgoal_tac "((fst m',0),n): (edges b)^+")
      prefer 2
      apply(blast intro:transitive_closure_trans )
      apply(blast dest:trancl_into_rtrancl)
      done

    from a1 and M0 and M2 
    have N3:"(fst m',1)\<in> fst b \<and> ((fst m',1),m'): (edges b)^+" 
          apply - 
      apply (simp only:mem_Collect_eq,(erule conjE)+,
         fold strand_def index_def nodes_def,
        drule_tac i="1" and ?n2.0="m'" in bundle_casual2_closed3)
      apply( simp, arith) 
      apply( subgoal_tac "((strand m', 1), m') \<in> (edges b )^*" )
      prefer 2 
      apply(blast intro:rtrancl_mono1) 
      apply(thin_tac "(m', n) \<in> (edges b)\<^sup>*")
          apply(drule_tac R="edges b" in  rtranclD)
      apply (erule disjE)   
      apply( simp add:node_equal)
      apply(unfold strand_def index_def)
          apply auto
      done 

    with M2 have N3a:"((fst m',1),n): (edges b)^*"
      apply -
      apply(subgoal_tac "((fst m',1),n): (edges b)^+")
      prefer 2
      apply(blast intro:transitive_closure_trans )
      apply(blast dest:trancl_into_rtrancl)
      done
          
    
        have "?t \<noteq> {|g0,h0|}" by simp
    
        from this  and M01 and M2 have "?t\<sqsubset> g0 | ?t \<sqsubset> h0"
          apply-  
      apply (simp  only:mem_Collect_eq)
      apply(erule conjE)+ 
      apply(drule  part_MPair)
      apply( auto)  done

        moreover
        {assume a14:"?t \<sqsubset>  g0" 
          
          from M01 and a14  have N4:" ?t \<sqsubset> node_term (fst m', 0)" by simp

          from this and M2  and N2 N2a  
      have  False 
        apply - 
        apply(unfold regularK_def Is_regular_strand_def 
          nodes_def strand_def)
        apply(erule conjE)+
        apply(drule_tac x="(fst m',0)" in spec )
        apply auto
        done
      
          from this have ?thesis by blast
        }
        moreover 
        {assume a13:" ?t \<sqsubset>  h0"
          
          from M01 and a13 have N4:"?t \<sqsubset> node_term ( fst m', 1)" by simp

          from this and M2 and  N3  N3a
      have  False
        apply - 
        apply(unfold regularK_def Is_regular_strand_def 
          nodes_def strand_def)
        apply(erule conjE)+
        apply(drule_tac x="(fst m',1)" in spec )
        apply auto
        done

          from this have ?thesis by blast
        }
        ultimately  have ?thesis by blast    
      }

      moreover
      {assume a12:"Is_Sep_strand (strand m')"
        from M3 and a8 and  a12 
    have M0:"snd m'=2|snd m'=1" 
          apply - 
      apply( unfold Is_Sep_strand_def, unfold Domain_def,
        unfold node_sign_def strand_def index_def,auto)
          apply(case_tac "y=0")
      apply simp 
      apply (case_tac "y=1" ) 
      apply simp 
      apply auto   done  
        from a12 
    have M00:"node_sign (fst m', 0)=-\<and> 
      node_sign (fst m', 1)=+\<and> 
      node_sign (fst m', 2)=+" 
          apply - 
      apply(unfold Is_Sep_strand_def,
        unfold  node_sign_def strand_def index_def, auto)
      done
        from a12 and this 
    have M1:"\<exists> g h . node_term (fst m',0)
      ={|g,h|} \<and> node_term (fst m', 1)=g \<and> 
      node_term (fst m',2)=h" (is "\<exists> g h. ?P g h")
          by (unfold node_term_def, unfold Is_Sep_strand_def 
        strand_def index_def, auto)
    then obtain g0 and h0 where M01:" ?P g0 h0"  by auto 
        from M0  have M0:"snd m'=2|snd m'=1" by simp
        moreover
        {assume a20:"snd m'=2"
          from a20  and a8 
      have N000:"(fst m',1):Domain\<and>((fst m', 1) ,m'):casual2^+"
            apply- apply( rule casual2_trancl)
        apply simp 
        apply simp done


          from a1 and M0 and M2 a20
      have N2:"(fst m',0): fst b\<and> ((fst m',0),m'): (edges b)^+" 
            apply - 
        apply (simp only:mem_Collect_eq,(erule conjE)+,
          fold strand_def index_def nodes_def,
          drule_tac i="0" and ?n2.0="m'" in bundle_casual2_closed3)
            apply( simp, arith) 
        apply( subgoal_tac "((strand m', 0), m') \<in> (edges b )^*" )
        prefer 2 
        apply(blast intro:rtrancl_mono1) 
        apply(thin_tac "(m', n) \<in> (edges b)\<^sup>*")
            apply(drule_tac R="edges b" in  rtranclD)
        apply (erule disjE)   
        apply( simp add:node_equal)
        apply(unfold strand_def index_def)
            apply auto
        done 

      with M2 have N2a:"((fst m',0),n): (edges b)^*"
        apply -
        apply(subgoal_tac "((fst m',0),n): (edges b)^+")
        prefer 2
        apply(blast intro:transitive_closure_trans )
        apply(blast dest:trancl_into_rtrancl)
        done
       

          from M2 and M01 and a20 have "?t \<sqsubset> h0" 
        by (subgoal_tac "(fst m',2)=m'",simp, simp add:node_equal,
          unfold strand_def index_def,auto)
           
          from this and  M2  and M01 and a20 
      have "?t \<sqsubset> node_term (fst m',0)" 
            apply - apply(simp) 
        apply( subgoal_tac "h0 \<sqsubset> node_term (fst m',0)") thm subterm_trans
            apply(drule_tac g="Crypt k h" and l=" node_term (fst m', 0)" in subterm_trans)
        apply(erule conjE)+ apply simp+
            apply(unfold parts_relation_def,  
          simp add:parts_insert_MPair parts_insert2) 
        apply blast 
            done

          from this and M2  and N2 N2a 
      have  False 
        apply - 
        apply(unfold regularK_def Is_regular_strand_def 
          nodes_def strand_def)
        apply(erule conjE)+
        apply(drule_tac x="(fst m',0)" in spec )
        apply auto
        done          
      
          from this have ?thesis by blast
        }
        moreover
        {assume a20:"snd m'=1"
          
          from a20  and a8 
      have N000:"(fst m',0):Domain\<and>((fst m', 0) ,m'):casual2^+"
            apply- apply( rule casual2_trancl)
        apply simp 
        apply simp done


          from a1 and M0 and M2 a20
      have N2:"(fst m',0): fst b\<and> ((fst m',0),m'): (edges b)^+" 
            apply - 
        apply (simp only:mem_Collect_eq,(erule conjE)+,
          fold strand_def index_def nodes_def,
          drule_tac i="0" and ?n2.0="m'" in bundle_casual2_closed3)
            apply( simp, arith) 
        apply( subgoal_tac "((strand m', 0), m') \<in> (edges b )^*" )
        prefer 2 
        apply(blast intro:rtrancl_mono1) 
        apply(thin_tac "(m', n) \<in> (edges b)\<^sup>*")
            apply(drule_tac R="edges b" in  rtranclD)
        apply simp
        apply (erule disjE)   
        apply( simp add:node_equal)
        apply(unfold strand_def index_def)
            apply auto
        done 

      with M2 have N2a:"((fst m',0),n): (edges b)^*"
        apply -
        apply(subgoal_tac "((fst m',0),n): (edges b)^+")
        prefer 2
        apply(blast intro:transitive_closure_trans )
        apply(blast dest:trancl_into_rtrancl)
        done
       

          from M2 and M01 and a20 have "?t \<sqsubset> g0" 
        apply -
        apply(subgoal_tac "(fst m',1)=m'",simp, simp add:node_equal,
          unfold strand_def index_def,auto)
        done
           
          

          from this  and M2  and M01 and a20 
      have "?t \<sqsubset> node_term (fst m',0)" 
            apply - apply(simp)
        apply( subgoal_tac "g0 \<sqsubset> node_term (fst m',0)") thm subterm_trans
            apply(drule_tac g="Crypt k h" and l=" node_term (fst m', 0)" in subterm_trans)
        apply(erule conjE)+ apply simp+
            apply(unfold parts_relation_def, 
          simp add:parts_insert_MPair parts_insert2) apply blast 
            done
      
          from this and M2  and N2  N2a have  False 
        apply - 
        apply(unfold regularK_def Is_regular_strand_def 
          nodes_def strand_def)
        apply(erule conjE)+
        apply(drule_tac x="(fst m',0)" in spec )
        apply auto
        done          
      
          from this have ?thesis by blast 
        }
        ultimately have ?thesis by blast  
      }

      moreover
      {assume a9:"Is_Tee_strand ?s"
    have N0:"(fst m',0): fst b\<and> ((fst m',0),m'): (edges b)^+"
    proof -
          from M3 and a8 and a9  have M0: "snd m' =2 | snd m'=1" 
        apply - apply( unfold Is_Tee_strand_def, unfold Domain_def 
          strand_def index_def, 
          auto)
            apply(case_tac "y=0") 
        apply(unfold node_sign_def, simp)
        apply auto done

          moreover
          {assume a10:"snd m'=2"
            from a1 and M0 and M2 and a10 have ?thesis 
              apply - 
          apply (simp only:mem_Collect_eq,(erule conjE)+,
         fold strand_def index_def nodes_def,
        drule_tac i="0" and ?n2.0="m'" in bundle_casual2_closed3)
              apply( simp, arith) 
          apply( subgoal_tac "((strand m', 0), m') \<in> (edges b )^*" )
          prefer 2 
          apply(blast intro:rtrancl_mono1) 
          apply(thin_tac "(m', n) \<in> (edges b)\<^sup>*")
              apply(drule_tac R="edges b" in  rtranclD)
          apply simp
          apply (erule disjE)   
          apply( simp add:node_equal)
          apply(unfold strand_def index_def)
              apply auto
          done 
          
          }
          
          moreover
          {assume a10:"snd m'=1"
        from a1 and M0 and M2 a10 have ?thesis 
              apply - 
          apply (simp only:mem_Collect_eq,(erule conjE)+,
        fold strand_def index_def nodes_def,
        drule_tac i="0" and ?n2.0="m'" in bundle_casual2_closed3)
              apply( simp, arith) 
          apply( subgoal_tac "((strand m', 0), m') \<in> (edges b )^*" )
          prefer 2 
          apply(blast intro:rtrancl_mono1) 
          apply(thin_tac "(m', n) \<in> (edges b)\<^sup>*")
              apply(drule_tac R="edges b" in  rtranclD)
          apply simp
          apply (erule disjE)   
          apply( simp add:node_equal)
          apply(unfold strand_def index_def)
              apply auto
          done 
            
          }
          ultimately show ?thesis by blast
    qed

    with M2 have N2a:"((fst m',0),n): (edges b)^*"
        apply -
        apply(subgoal_tac "((fst m',0),n): (edges b)^+")
        prefer 2
        apply(blast intro:transitive_closure_trans )
        apply(blast dest:trancl_into_rtrancl)
        done
        
        from M3 and a8 and a9  
    have M0: "snd m' =2 | snd m'=1" 
      apply - 
      apply( unfold Is_Tee_strand_def, unfold Domain_def 
        strand_def index_def , auto)
          apply(case_tac "y=0")
      apply(unfold node_sign_def, simp) 
      apply auto done

        from a9 and this  
    have "\<exists> g . node_term (fst m',0)=g \<and>
      node_term (fst m', 1)=g \<and> 
      node_term m'=g" (is "\<exists> g. ?P g")
          apply - 
      apply (unfold node_term_def, unfold Is_Tee_strand_def
        strand_def index_def , auto) done
        then obtain g where M01:" ?P g" ..

        from M2 and this have "?t \<sqsubset> node_term (fst m', 0)" by simp
        
        from this and N0 and M2 N2a
    have False
      apply - 
      apply(unfold regularK_def Is_regular_strand_def 
        nodes_def strand_def)
      apply(erule conjE)+
      apply(drule_tac x="(fst m',0)" in spec )
      apply auto
      done       
    
        then  have  ?thesis by blast
      }
      
      moreover
      {assume a9:"Is_T_strand ?s"
    from M3 and a9 and a8  have N0:"snd m'=0" 
      apply -
          apply (unfold Is_T_strand_def, unfold node_sign_def
        strand_def index_def, unfold Domain_def ) 
      apply auto done

    from a9 and this 
    have "\<exists> g. g \<in> T  \<and>  node_term (fst m',0)=g " (is "\<exists> g. g \<in> T \<and>  ?P g")
          apply - apply (unfold node_term_def, unfold
        Is_T_strand_def strand_def index_def, 
        auto) 
      done

    then obtain g where M01:" g:T\<and>?P g" by blast  

    from N0 and M2 and this  have "?t \<sqsubset> g" 
      apply - 
      apply (subgoal_tac "(fst m',0)=m'",
        simp, simp add:node_equal,
      unfold strand_def index_def,auto) done

    from this and M01 have ?thesis 
      apply - 
      apply (simp only:T_is_only_text) 
      apply(unfold Atoms_def,
        unfold  parts_relation_def,auto) done
      } 
      moreover
      {assume a10 :"Is_K_strand (strand m')"
    from M3 and a10 and a8  have N0:"snd m'=0"
      apply -
          apply (unfold Is_K_strand_def, 
        unfold node_sign_def strand_def index_def, 
        unfold Domain_def strand_def  index_def) 
      apply auto
      done

    from a10 and this 
    have "\<exists> k. k\<in> KP \<and>  node_term (fst m',0)=Key k " (is "\<exists> k. k\<in> KP \<and> ?P k")
          apply - 
      apply (unfold node_term_def, unfold Is_K_strand_def
        index_def strand_def , auto) 
      done

    then obtain k0 where M01:" k0:KP\<and>?P k0" 
      by blast  

    from N0 and M2 and this  have "?t \<sqsubset> Key k0" 
      apply - apply (subgoal_tac "(fst m',0)=m'",
        simp, simp add:node_equal,
        unfold strand_def  index_def,
        auto) 
      done

    from this and M01 
    have ?thesis 
      apply -  
      apply(unfold  parts_relation_def,auto) done   
      } 
      moreover
      { assume a11:"Is_Flush_strand (strand  m')"
    from M3 and a11 and a8 have ?thesis
      apply - 
      apply(unfold Is_Flush_strand_def,unfold Domain_def,
        unfold node_sign_def,
        unfold Domain_def strand_def  index_def,auto) 
      done
      }


      
      ultimately show  ?thesis by blast

    qed
  qed

  from a1 and M2 and M3 and this 
  show ?thesis 
    apply - 
    apply(rule_tac x="m'" in exI )
    apply(rule conjI,simp)+
    apply(rule allI)+
    apply(erule conjE)+
    apply(drule_tac x="y" in spec)
    apply(rule impI) thm bundle_is_trans_closed22
    apply(subgoal_tac "(y,n):(edges b)^*")
    apply(drule bundle_is_trans_closed22 )
    apply simp  apply simp thm transitive_closure_trans
    apply(subgoal_tac "(y,m'):(edges b)^*")
    apply(blast intro:transitive_closure_trans )
    apply(blast dest:trancl_into_rtrancl)
   
    done


qed

    
lemma in_auth_test: 
  assumes a1:"b:bundles" and  a2:" unique_originate a n" 
  and a3:"suite M a b" 
  and a4:"n':nodes b"
  and a5:"a\<sqsubset>h"
  and a6:"Crypt k h \<sqsubset> node_term n'"
  and a8:"\<not>Crypt k h \<sqsubset> node_term n"
  and a9:"regularK k b"

  shows
  "\<exists>  m' m.((n,m):(edges b)^* \<and> (m',n'):(edges b)^*\<and> m\<Rightarrow>\<^sup>+m'  \<and> node_sign m'=+ 
              \<and>( strand m'\<noteq>strand n\<longrightarrow>  node_sign m=- )
              \<and> \<not> Crypt k h \<sqsubset> node_term m \<and> Crypt k h \<sqsubset> node_term m'
              \<and>  \<not> Crypt k h \<sqsubset>  node_term m \<and>  Crypt k h \<sqsubset> node_term m'
              \<and>Is_regular_strand (strand m'))
  "
  
 
proof -
  have b1:"\<exists> m'. (m',n'):(edges b)^* \<and>m': (nodes b) \<and> 
    Crypt k h \<sqsubset> node_term m'\<and> 
    Is_regular_strand (strand m')
    \<and> node_sign m'=+ \<and>
    (\<forall> y.(y,m'):(edges b)^+ \<longrightarrow>\<not> Crypt k h \<sqsubset> node_term y)" (is "\<exists> m'. ?P m'")
    by(rule unsolicited_test)

  then obtain m' where b1:"?P m'" by blast

  

  
  

 from b1 a5 have M4:"a \<sqsubset> node_term m'" thm subterm_trans
   apply(rule_tac h="h" in subterm_trans)
   apply auto
   apply(rule_tac h="Crypt k h" in subterm_trans)
   apply auto
   apply(unfold parts_relation_def,auto)
   done

 from b1 a8 have b2:"m'\<noteq>n"
    apply - apply(rule ccontr,auto)
    done

 from this and a2
  have "strand n= strand m' \<and> (a \<sqsubset> node_term m' \<longrightarrow> index n < index m') |
    strand n\<noteq> strand m' \<and> non_originate a (strand m')"
    apply - apply(rule  non_originate_node)
    by auto

  moreover
  {assume M9:" strand n\<noteq> strand m' \<and> non_originate a (strand m')"
    thm non_originate_node1
    have "\<exists>m. first_node_in_nonorigi_strand a m' m" 
      apply - apply(cut_tac a2 M9 M4)
      apply(blast intro!:non_originate_node1)
      done

    then obtain m where M10:"first_node_in_nonorigi_strand a m' m"
      by blast

    then have M11:"a \<sqsubset> node_term m"
      by(unfold first_node_in_nonorigi_strand_def,auto)

    from M10 M4 have M12:"index m <=index m'"
      apply - apply(rule nonOrigNode1)
      apply auto
      done
    thm bundle_casual2_closed3
    have "(strand m', index m) \<in> nodes b \<and> 
      ((strand m', index m), m') \<in> (edges b \<inter> casual2)\<^sup>*"
      apply - apply(cut_tac a1 M12 b1)
      apply(rule  bundle_casual2_closed3)
      by auto

    with M10 have M13:"m:nodes b\<and>(m,m'):(edges b \<inter> casual2)\<^sup>*"
      apply - 
      apply(subgoal_tac "(strand m', index m)=m",simp)
      apply(unfold first_node_in_nonorigi_strand_def)
      apply(simp add:node_equal)
      apply(unfold strand_def index_def,auto)
      done

   then  have M14:"(m,m'):( casual2)\<^sup>*" 
      apply - thm  rtrancl_mono 
      apply(subgoal_tac "(edges b \<inter> casual2)^*\<subseteq> casual2^*")
      apply (blast intro!: rtrancl_mono)+done

    
    have M15: "m\<noteq>m'"
      apply - apply(rule ccontr)
      apply(cut_tac b1 M10)
      apply(unfold first_node_in_nonorigi_strand_def, auto)
      done

    with M14 have M16:"(m,m'):( casual2)\<^sup>+" 
      apply - apply(blast dest:rtranclD )
      done

    from M13 have M17:"(m,m'):(edges b)\<^sup>*" 
      apply - thm  rtrancl_mono 
      apply(subgoal_tac "(edges b \<inter> casual2)^*\<subseteq> (edges b)^*")
      apply (blast intro!: rtrancl_mono)+done

    
    with M15 have M18:"(m,m'):( edges b)\<^sup>+" 
      apply - apply(blast dest:rtranclD )
      done
    
    from M17  b1 M14 have M19:"(m,n'):(edges b)^*"
      by(blast dest:transitive_closure_trans)

    from M19 M18 M11 b1
    have M20:"\<not> Crypt k h \<sqsubset> node_term m"
      apply -
      apply(erule conjE)+
      apply(drule_tac x="m" in spec)
      by blast

    from M16 have M21:"\<exists> m.(m,m'):casual2^+" 
      by blast 
    
    have M22:"Is_regular_strand (strand m')"
      apply(cut_tac b1) apply blast
      done

    
    from M11  and a1 and a2 and M13
    have M23: "(\<exists> p.  p: (complete_Path a b)  \<and> (last p)=m)" (is "\<exists> p. ?R p")
      thm path_exist
      by(blast dest:path_exist)

    from this obtain p where "?R p" by auto
    from a1 this a2 M13 have M24: "n=m|(n,m):(edges b)^+" thm completePathProp
      apply - apply((erule conjE), drule completePathProp)
      apply blast
      done

    from this have M24:"(n,m):(edges b)^*"  by (blast dest:trancl_into_rtrancl)
    
    have ?thesis 
      apply - 
      apply(rule_tac x="m'" in exI) 
      apply(rule_tac x="m" in exI) 
      apply(cut_tac M22 M20 M19 M16 b1 M10 M24 M10) 
      apply(unfold first_node_in_nonorigi_strand_def)
      apply(rule conjI, blast)+  
      
      apply auto
      done
  }
  moreover
  { assume
    M9:"strand n= strand m' \<and> (a \<sqsubset> node_term m' \<longrightarrow> index n < index m') "
    from M4 M9
    have M10:"index n < index m'"
      by simp

    from this  have M12:"index n <=index m'"
      by simp
    thm bundle_casual2_closed3
    have "(strand m', index n) \<in> nodes b \<and> 
      ((strand m', index n), m') \<in> (edges b \<inter> casual2)\<^sup>*"
      apply - apply(cut_tac a1 M12 b1)
      apply(rule  bundle_casual2_closed3)
      by auto

    with M9 have M13:"n\<in> nodes b\<and>(n,m'):(edges b \<inter> casual2)\<^sup>*"
      apply - 
      apply(subgoal_tac "(strand m', index n)=n",simp)
      
      apply(simp add:node_equal)
      apply(unfold strand_def index_def,auto)
      done

   then  have M14:"(n,m'):( casual2)\<^sup>*" 
      apply - thm  rtrancl_mono 
      apply(subgoal_tac "(edges b \<inter> casual2)^*\<subseteq> casual2^*")
      apply (blast intro!: rtrancl_mono)+done

    
    have M15: "n\<noteq>m'"
      apply - apply(rule ccontr)
      apply(cut_tac b1 M10)
      apply( auto)
      done

    with M14 have M16:"(n,m'):( casual2)\<^sup>+" 
      apply - apply(blast dest:rtranclD )
      done

    then have M17:"\<exists> m.(m,m'):( casual2)\<^sup>+" 
      by blast
    
    have M22:"Is_regular_strand (strand m')"
      apply(cut_tac b1) apply blast
      
      done

    
    have ?thesis 
      apply - 
      apply(rule_tac x="m'" in exI) 
      apply(rule_tac x="n" in exI) 

      apply(cut_tac M22  M16 b1 M9 M10 a2 a8) 
      apply(unfold unique_originate_def originate_def)
      apply(rule conjI,simp)+
      
      apply auto
      done
  }
  ultimately show ?thesis by blast
      
qed

 

end
